####
# bomail.cli.send
#
# Send drafts.
####

import os
import sys
import mimetypes
import subprocess, email
import smtplib

from email import encoders
from email.mime.text import MIMEText
from email.mime.multipart import MIMEMultipart
from email.mime.audio import MIMEAudio
from email.mime.base import MIMEBase
from email.mime.image import MIMEImage

import bomail.config.config as config
import bomail.cli.mailfile as mailfile
import bomail.util.addr as addr

usage_str = """
Reads filename(s) from stdin one per line and sends them.
send.py -h to print this help.
"""


# adapted from python library example: https://docs.python.org/3/library/email-examples.html
def add_attachments(msg, attach_list):
  for filename in attach_list:
    if not os.path.exists(filename):
      raise Exception("Did not send. Could not attach (does not exist): " + filename)
    if not os.path.isfile(filename):
      raise Exception("Did not send. Could not attach (is not a file): " + filename)
    ctype, encoding = mimetypes.guess_type(filename)
    if ctype is None or encoding is not None:
      ctype = 'application/octet-stream'
    maintype, subtype = ctype.split('/', 1)
    if maintype == 'text':
      with open(filename) as fp:
        sub_msg = MIMEText(fp.read(), _subtype=subtype)
    elif maintype == 'image':
      with open(filename, 'rb') as fp:
        sub_msg = MIMEImage(fp.read(), _subtype=subtype)
    elif maintype == 'audio':
      with open(filename, 'rb') as fp:
        sub_msg = MIMEAudio(fp.read(), _subtype=subtype)
    else:
      with open(filename, 'rb') as fp:
        sub_msg = MIMEBase(maintype, subtype)
        sub_msg.set_payload(fp.read())
      encoders.encode_base64(sub_msg)
    sub_msg.add_header('Content-Disposition', 'attachment', filename=os.path.split(filename)[1])
    msg.attach(sub_msg)


# return num_successes, err_msg
def main(filelist, mgr):
  try:
    serv = smtplib.SMTP(config.smtp_servername, config.smtp_port)
    serv.ehlo()
    serv.starttls()
    serv.ehlo()
    serv.login(config.email_addr, config.password)
  except Exception as e:
    return 0, "Error connecting to server: " + str(e)

  for i, filename in enumerate(filelist):
    data = mgr.get_all(filename)
    attach_s = data[mailfile.ATTACH_L].strip()
    if len(attach_s) == 0:  # no attachments
      msg = MIMEText(data[mailfile.BODY_L], 'plain')
    else:
      msg = MIMEMultipart()
      msg.attach(MIMEText(data[mailfile.BODY_L], 'plain'))
      add_attachments(msg, attach_s.split(", "))
    msg['Subject'] = data[mailfile.SUBJ_L]
    msg['From'] = data[mailfile.FROM_L]
    msg['To'] = data[mailfile.TO_L]
    msg['CC'] = data[mailfile.CC_L]
    refs = data[mailfile.REFS_L]
    if len(refs) > 0:
      msg['In-Reply-To'] = refs.split(", ")[-1]
      msg['References'] = refs
    recip_lists = data[mailfile.TO_L].split(", ")
    if len(data[mailfile.CC_L]) > 0:
      recip_lists += data[mailfile.CC_L].split(", ")
    if len(data[mailfile.BCC_L]) > 0:
      recip_lists += data[mailfile.BCC_L].split(", ")
    recip_addrs = [addr.str_to_pair(a)[1] for a in recip_lists]

    try:
      serv.sendmail(config.email_addr, recip_addrs, msg.as_string())
    except Exception as e:
      serv.quit()
      return i, "Error sending message #" + str(i) + " [" + filename + "]: " + str(e)

  serv.quit()
  return len(filelist), ""


def main_cli():
  if len(sys.argv) >= 2:
    print(usage_str)
    exit(0)

  filelist = [f.strip() for f in sys.stdin.readlines()]
  res, err = main(filelist, mailfile.MailMgr())
  print("Sent " + str(res) + "/" + str(len(filelist)) + " messages.")
  if res < len(filelist):
    print(err)


if __name__ == "__main__":
  main_cli()





####
# bomail.util.search_opts
#
# The main 'search' code.
####

import os
import shlex

import bomail.cli.mailfile as mailfile
import bomail.util.tags as tags
import bomail.util.datestr as datestr

####
# The code tries to be a bit fancy -- given a search string,
# generate and compile a short python program that conducts
# the search using only the options in the query.
####

options_str = """
All arguments are optional:
 -h               [print this help]

 -not             not satisfying the next option
 -OR              matching everything before OR everything after
 
 -a datestr       after date. datestr can be e.g. yyyy-mm-dd
 -b datestr       before date

 -open            state is open
 -scheduled       state is scheduled
 -closed          state is closed
 -draft           is a draft
 -sent            is sent from me
 -attach          has an attachment

 -t str           has str as a tag. Use multiple times to match any
 -t "tag1, tag2"  has all these tags
 -nt str          does not have str as a tag
 -nt "tag1, tag2" does not have *all* of these tags
 -notags          has no tags

 -to str          str is in to, cc, or bcc field
 -to "s1, s2"     s1, s2, ... all in to, cc, or bcc field
 -from str        str is in from field
 -subject str     str is in subject field

 query            query is in email somewhere
 "long query"     long query is in email somewhere
 quer1 quer2      all queries are in email somewhere

 -n 200           only return first 200 results
 -sortold         list by oldest (default is by newest)
"""


# used when constructing the search code
break_str = "      break\n"


# dumb helper function: when producing python code from an object
# that is either None or a string
def mystr(s):
  if s is None:
    return "None"
  else:
    return "\"" + s + "\""


# filename is ...email/yyyy/mm-dd/x.email
# or .../yyyy/mm-dd/x.draft
def get_date_from_filename(filename):
  # filedir = os.path.dirname(filename)  # too slow
  slashind = filename.rfind("/")  # assume x does not contain a slash!!
  return filename[slashind-10:slashind].replace("/","-")


def datestr_matches(datestr, after, before, not_after=None, not_before=None):
  return ((after is None or datestr[:len(after)] >= after)
          and (before is None or datestr[:len(before)] <= before)
          and (not_after is None or datestr < not_after)
          and (not_before is None or datestr > not_before))


# given a 'datestr' like yyyy-mm or m3.0d,
# turn it into an absolute date string yyyy-mm-dd
# throw an exception if unable
def get_absstr_from_datestr(schedstr):
  dateobj = datestr.get_datetime(schedstr)
  if dateobj is None:
    dateobj = datetime.datetim.now()  # what else to do?
  return dateobj.isoformat()[:10]


# A query that does not contain an OR
class OneQuery:
  def __init__(self, copy=None):
    if copy is None:
      self.after = None
      self.not_after = None
      self.before = None
      self.not_before = None
  
      self.open = None
      self.scheduled = None
      self.closed = None
      self.draft_only = None
      self.sent_only = None
      self.attachonly = None
  
      self.tagset_list = None
      self.nottagset_list = None
      self.notags = None  # or bool
  
      self.tolist = None
      self.not_tolist = None
      self.fromlist = None
      self.not_fromlist = None
  
      self.subject_str = None
      self.not_subject_str = None
  
      self.querylist = None
      self.not_querylist = None
  
      self.need_data = False
      self.need_date = False
    else:
      self.after = copy.after
      self.not_after = copy.not_after
      self.before = copy.before
      self.not_before = copy.not_before
  
      self.open = copy.open
      self.scheduled = copy.scheduled
      self.closed = copy.closed
      self.draft_only = copy.draft_only
      self.sent_only = copy.sent_only
      self.attachonly = copy.attachonly
  
      self.tagset_list = copy.tagset_list
      self.nottagset_list = copy.nottagset_list
      self.notags = copy.notags
  
      self.tolist = copy.tolist
      self.not_tolist = copy.not_tolist
      self.fromlist = copy.fromlist
      self.not_fromlist = copy.not_fromlist
  
      self.subject_str = copy.subject_str
      self.not_subject_str = copy.not_subject_str
  
      self.querylist = copy.querylist
      self.not_querylist = copy.not_querylist
  
      self.need_data = copy.need_data
      self.need_date = copy.need_date
      

  def parse(self, args):
    i = 0
    while i < len(args):
      same_as_current = True
      if args[i] == "-nt":
        # "hack" to make this equivalent to -not -t
        same_as_current = False
        args[i] = "-t"
      
      # NOT elif!
      if args[i] == "-not":
        same_as_current = False
        i += 1
        if i >= len(args):
          break
      
      if args[i] == "-a":
        if i+1 < len(args):
          try:
            self.need_date = True
            if same_as_current:
              res = get_absstr_from_datestr(args[i+1])
              self.after = res
            else:
              res = get_absstr_from_datestr(args[i+1])
              self.not_after = res
          except:
            self.need_date = False
            raise
        i += 1
      elif args[i] == "-b":
        if i+1 < len(args):
          try:
            self.need_date = True
            if same_as_current:
              res = get_absstr_from_datestr(args[i+1])
              self.before = res
            else:
              res = get_absstr_from_datestr(args[i+1])
              self.not_before = res
          except:
            self.need_date = False
            raise
        i += 1

      elif args[i] == "-open":
        self.need_data = True
        self.open = same_as_current
      elif args[i] == "-closed":
        self.need_data = True
        self.closed = same_as_current
      elif args[i] == "-scheduled":
        self.need_data = True
        self.scheduled = same_as_current
      elif args[i] == "-draft":
        self.need_data = True
        self.draft_only = same_as_current
      elif args[i] == "-sent":
        self.need_data = True
        self.sent_only = same_as_current
      elif args[i] == "-attach":
        self.need_data = True
        self.attachonly = same_as_current

      elif args[i] == "-t":
        if i+1 < len(args):
          # don't split query tags into folders, only search for the exact query
          the_tagset = tags.get_tagset_from_str(args[i+1], include_folders=False)
          if len(the_tagset) > 0:
            self.need_data = True
            if same_as_current:
              if self.tagset_list is None:
                self.tagset_list = [the_tagset]
              else:
                self.tagset_list.append(the_tagset)
            else:
              if self.nottagset_list is None:
                self.nottagset_list = [the_tagset]
              else:
                self.nottagset_list.append(the_tagset)
        i += 1
      elif args[i] == "-notags":
        self.need_data = True
        self.notags = same_as_current

      elif args[i] == "-to":
        if i+1 < len(args):
          self.need_data = True
          the_tolist = [s.strip() for s in args[i+1].split(",")]
          if same_as_current:
            self.tolist = the_tolist
          else:
            self.not_tolist = the_tolist
        i += 1
      elif args[i] == "-from":
        if i+1 < len(args):
          self.need_data = True
          the_fromlist = [s.strip() for s in args[i+1].split(",")]
          if same_as_current:
            self.fromlist = the_fromlist
          else:
            self.not_fromlist = the_fromlist
        i += 1

      elif args[i] == "-subject":
        if i+1 < len(args):
          self.need_data = True
          if same_as_current:
            self.subject_str = args[i+1]
          else:
            self.not_subject_str = args[i+1]
        i += 1

      else:
        self.need_data = True
        if same_as_current:
          if self.querylist is None:
            self.querylist = []
          self.querylist.append(args[i])
        else:
          if self.not_querylist is None:
            self.not_querylist = []
          self.not_querylist.append(args[i])

      i += 1

  # append a list of strings of commands to slist
  # a 4-indented block that checks f based on self, mailfile, mgr, tags, data (if needed), datestr (if needed), datestr_matches (if needed)
  # if a check fails, break
  def compile_slist(self, slist):
    if self.draft_only is not None:
      slist.append("    if f[-5:] ")
      slist.append("!" if self.draft_only else "=")
      slist.append("= \"draft\":\n")
      slist.append(break_str)
    if any([o is not None for o in [self.after, self.not_after, self.before, self.not_before]]):
      slist += ["    if not datestr_matches(datestr, ", mystr(self.after), ", ", mystr(self.before), ", ", mystr(self.not_after), ", ", mystr(self.not_before), "):\n"]
      slist.append(break_str)

    if self.open is not None:
      slist.append("    if data[mailfile.STATE_L].startswith(\"open\") != ")
      slist.append(str(self.open))
      slist.append(":\n")
      slist.append(break_str)
    if self.closed is not None:
      slist.append("    if data[mailfile.STATE_L].startswith(\"closed\") != ")
      slist.append(str(self.closed))
      slist.append(":\n")
      slist.append(break_str)
    if self.scheduled is not None:
      slist.append("    if data[mailfile.STATE_L].startswith(\"scheduled\") != ")
      slist.append(str(self.scheduled))
      slist.append(":\n")
      slist.append(break_str)
    if self.sent_only is not None:
      slist.append("    if (data[mailfile.SENT_L] == \"True\") != ")
      slist.append(str(self.sent_only))
      slist.append(":\n")
      slist.append(break_str)
    if self.attachonly is not None:
      slist.append("    if (len(data[mailfile.ATTACH_L]) > 0) != ")
      slist.append(str(self.attachonly))
      slist.append(":\n")
      slist.append(break_str)

    if self.notags is not None:
      slist.append("    if len(data[mailfile.TAGS_L]) ")
      slist.append("!" if self.notags else "=")
      slist.append("= 0:\n")
      slist.append(break_str)

    # TODO: speed these up by manually embedding the tags we've already parsed? maybe not, set is fast lookup..
    got_tagset = False
    if self.tagset_list is not None:
      # if the search includes a folder, make sure to match it
      slist.append("    data_tagset = tags.get_tagset_from_str(data[mailfile.TAGS_L], include_folders=True)\n")
      slist.append("    if not any([ts.issubset(data_tagset) for ts in q.tagset_list]):\n")
      slist.append(break_str)
      got_tagset = True
    if self.nottagset_list is not None:
      if not got_tagset:
        slist.append("    data_tagset = tags.get_tagset_from_str(data[mailfile.TAGS_L], include_folders=True)\n")
        got_tagset = True
      slist.append("    if any([nts.issubset(data_tagset) for nts in q.nottagset_list]):\n")
      slist.append(break_str)

    if self.tolist is not None:
      for tostr in self.tolist:
        dec = "\"" + tostr.encode().decode('unicode_escape') + "\""
        slist += ["    if ", dec, " not in data[mailfile.TO_L] and ", dec, " not in data[mailfile.CC_L] and ", dec, " not in data[mailfile.BCC_L]:\n", break_str]
    if self.not_tolist is not None:
      for tostr in self.not_tolist:
        dec = "\"" + tostr.encode().decode('unicode_escape') + "\""
        slist += ["    if ", dec, " in data[mailfile.TO_L] or ", dec, " in data[mailfile.CC_L] or ", dec, " in mailfile.BCC_L:\n"]
      slist.append(break_str)
    if self.fromlist is not None:
      for fromstr in self.fromlist:
        slist += ["    if \"", fromstr.encode().decode('unicode_escape'), "\" not in data[mailfile.FROM_L]:\n"]
      slist.append(break_str)
    if self.not_fromlist is not None:
      for fromstr in self.not_fromlist:
        slist += ["    if \"", fromstr.encode().decode('unicode_escape'), "\" in data[mailfile.FROM_L]:\n"]
        slist.append(break_str)

    if self.subject_str is not None:
      slist += ["    if \"", self.subject_str.encode().decode('unicode_escape'), "\" not in data[mailfile.SUBJ_L]:\n"]
      slist.append(break_str)
    if self.not_subject_str is not None:
      slist += ["    if \"", self.not_subject_str.encode().decode('unicode_escape'), "\" in data[mailfile.SUBJ_L]:\n"]
      slist.append(break_str)

    if self.querylist is not None:
      for q in self.querylist:
        slist += ["    if not any([\"", q.encode().decode('unicode_escape'), "\" in l for l in data]):\n"]
        slist.append(break_str)
    if self.not_querylist is not None:
      for q in self.not_querylist:
        slist += ["    if any([\"", q.encode().decode('unicode_escape'), "\" in l for l in data]):\n"]
        slist.append(break_str)


def list_indic(l):
  return False if len(l) == 0 else all(l)


def do_compile(q_list, need_data, need_date):
  slist = ["matched_list = []\nfor i,f in enumerate(filelist):\n"]
  if need_data:
    slist.append("  data = mgr.get_all(f) if datalist is None else datalist[i]\n")
    if need_date:
      slist.append("  datestr = data[mailfile.DATE_L]\n")
  elif need_date:
    slist.append('  slashind = f.rfind("/")\n  datestr = f[slashind-10:slashind].replace("/","-")\n')

  slist.append("  matched_result = False\n")
  for i,q in enumerate(q_list):
    slist.append("  q = self.q_list[" + str(i) + "]\n")
    # if any q sets matched_result True, it's a match (logical OR)
    slist.append("  while True:\n")
    # a 4-indented block that breaks if not a match
    q.compile_slist(slist)
    slist.append("    matched_result = True\n    break\n")  # if all tests pass, match!
    slist.append("  if matched_result:\n    matched_list.append(i)\n    continue\n")

  slist = slist
#  with open("compile_log.txt", "w") as f:
#    f.write("".join(slist))
  return compile("".join(slist), '<string>', 'exec')


# A full search query
class SearchQuery:
  # if include_listmailed, then we can assume listmail
  # has already filtered the list of filenames for us
  def __init__(self, arglist, include_listmailed=True):
    self.max_results = -1  # no limit
    self.sort_new = True
    self.q_list = []
    self.compiled_obj = None
    self.compiled_listmailed_obj = None
    self.need_data = False
    self.need_date = False

    if len(arglist) == 0:
      return

    if "-n" in arglist:
      ind = arglist.index("-n")
      if ind+1 < len(arglist):
        self.max_results = int(arglist[ind+1])
      arglist = arglist[:ind] + ([] if len(arglist) <= ind+2 else arglist[ind+2:])

    if "-sortold" in arglist:
      self.sort_new = False
      ind = arglist.index("-sortold")
      arglist = arglist[:ind] + ([] if len(arglist) <= ind+1 else arglist[ind+1:])

    i = 0
    while i < len(arglist) and "-OR" in arglist[i:]:
      or_i = arglist.index("-OR", i)
      q = OneQuery()
      q.parse(arglist[i:or_i])
      self.q_list.append(q)
      i = or_i + 1
    if i < len(arglist):
      q = OneQuery()
      q.parse(arglist[i:])
      self.q_list.append(q)

    if len(self.q_list) > 0:
      self.need_data = any([q.need_data for q in self.q_list])
      self.need_date = any([q.need_date for q in self.q_list])

    self.compiled_obj = do_compile(self.q_list, self.need_data, self.need_date)

    if include_listmailed:
      # TODO! (complicated)
      # produce the version that assumes listmail has been run before
      # calling this to filter it
      # important that order or draft, open, scheduled, attach
      # matches the order that listmail uses to break ties in which query to address
      # this could speed up some searches a lot...
      self.compiled_listmailed_obj = self.compiled_obj
        
        

  # return list of indices of matching files in filelist
  # if listmailed, don't need to repeat work...
  def filter(self, filelist, mgr, datalist=None, listmailed=False):
    if len(self.q_list) == 0:
      return range(len(filelist))
    namespace ={"self": self, "datestr_matches": datestr_matches, "mailfile": mailfile, "mgr": mgr, "tags": tags, "filelist": filelist, "datalist": datalist}
    if listmailed:
      exec(self.compiled_obj, namespace)
    else:
      exec(self.compiled_listmailed_obj, namespace)
    return namespace["matched_list"]


  # get options for listmail
  # it's ok to return too many results, but not to leave some out
  # (although the queries may conflict...)
  # after and before are None or strings
  # others are bool
  def get_list_opts(self):
    if len(self.q_list) == 0:
      return None, None, False, False, False, False

    after_list = [q.after for q in self.q_list if q.after is not None]
    after = None if len(after_list) == 0 else min(after_list)
    before_list = [q.before for q in self.q_list if q.before is not None]
    before = None if len(before_list) == 0 else max(before_list)

    draft = list_indic([q.draft_only for q in self.q_list if q.draft_only is not None])
    is_open = list_indic([q.open for q in self.q_list if q.open is not None])
    scheduled = list_indic([q.scheduled for q in self.q_list if q.scheduled is not None])
    attachonly = list_indic([q.attachonly for q in self.q_list if q.attachonly is not None])
    return after, before, draft, is_open, scheduled, attachonly


def SQ_from_str(s):
  return SearchQuery(shlex.split(s))





####
# bomail.cli.mailfile
#
# Specifies the bomail file format, called 'mailfile',
# and mail manager object, and some commands to view/edit
# metadata of emails.
####

import sys
import os, bisect
import email.utils, uuid

import datetime
import dateutil, dateutil.parser

import bomail.config.config as config
import bomail.util.tags as tags
import bomail.util.msgids as msgids

####
# This file specifies
# (1) the mailfile format and interface
# (2) MailMgr, the mail manager object

# Emails are stored in "mailfiles".
# A mailfile is of the form:
#   field_name: field_value
#   ...
#   field_name: field_value
#   =================
#   body_of_email

# Formats of specific entries:
# "addr" can be either of these:
#    name@domain.com
#    Longer Name Here <name@domain.com>
#
# "msg_id" is an email message id enclosed in < >
#
# "date" is an ISO8061 date string in UTC time, i.e. yyyy-mm-ddTHH:MM:SS or some prefix
#
# no newlines except in the body
#
# all lists are comma-and-space separated, implying that their
# elements shouldn't have either commas or spaces in them
####

fields = ["from","to","cc","bcc","subject","date","sent","msg-id","references","referenced_by","attachments","state","tags","body"]
FROM_L   = 0    # addr
TO_L     = 1    # addr_list
CC_L     = 2    # addr_list
BCC_L    = 3    # addr_list
SUBJ_L   = 4    # string with no newline
DATE_L   = 5    # datestr
SENT_L   = 6    # "True" if sent from me, "False" otherwise
MSG_ID_L = 7    # msg_id
REFS_L   = 8    # list of msg_ids
REF_BY_L = 9    # list of msg_ids
ATTACH_L = 10   # list of attachment files
STATE_L  = 11   # "open", "closed", or "scheduled datestr"
TAGS_L   = 12   # list of tags
BODY_L   = 13   # arbitrary string

sep = "================\n"  # between headers and body

usage_str = """
Usage: reads a list of filenames from stdin and does the action for each one.
Options: -g (get the indicated fields), -add-tags, -rm-tags

  mailfile.py -h                  # print this help
  mailfile.py -g all              # get and print all fields
  mailfile.py -g from bcc date    # get and print the fields 'from', 'bcc', and 'date'
  mailfile.py -add-tags tag1 tag2 # add the tags 'tag1' and 'tag2'
  mailfile.py -rm-tags tag1 tag2  # remove the tags 'tag1' and 'tag2'
"""


error_data = [
  "error file not found <>",
  "", "", "", "", "9999-12-28", "True", email.utils.make_msgid(uuid.uuid4().hex),
  "", "", "", "open", "", "error file not found"]


def msgid_as_path(msg_id):
#  return msg_id.replace("<","").replace(">","").replace("/","-")
  return msg_id.replace("/","-")


# input date in isoformat
# output directory, prefix
def get_datepath(datestr):
  s = dateutil.parser.parse(datestr).astimezone(datetime.timezone.utc).strftime("%Y/%m-%d/T%H:%M:%S")
  return s[:-9], s[-9:]


def wrong_get_datepath(datestr):
  s = dateutil.parser.parse(datestr).astimezone(datetime.timezone.utc).strftime("%Y/%m-%d/T%H:%S")
  return s[:-6], s[-6:]
def wrong_get_rel_idpath(datestr, msgid):
  mydir, ext = wrong_get_datepath(datestr)
  return mydir, ext + "-" + msgid_as_path(msgid)
def wrong_get_attachdir(data):
  mydir, ext = wrong_get_rel_idpath(data[DATE_L], data[MSG_ID_L])
  return config.attach_dir + mydir + ext + "/"


def get_rel_idpath(datestr, msgid):
  mydir, ext = get_datepath(datestr)
  return mydir, ext + "-" + msgid_as_path(msgid)


# assumes it's not a draft
# return dir, filename
def get_mailfilepath(data):
  mydir, ext = get_rel_idpath(data[DATE_L], data[MSG_ID_L])
  return config.email_dir + mydir, ext + ".mail"


def get_attach_dir(data):
  mydir, ext = get_rel_idpath(data[DATE_L], data[MSG_ID_L])
  return config.attach_dir + mydir + ext + "/"


def get_trash_attach_dir(filename, data):
  mydir, ext = get_rel_idpath(data[DATE_L], data[MSG_ID_L])
  return config.trash_dir + "attach/" + mydir + ext + "/"


def get_trashfilepath(filename, data):
  mydir, ext = get_rel_idpath(data[DATE_L], data[MSG_ID_L])
  return config.trash_dir + mydir, ext + filename[filename.rfind("."):]


def get_draftfilepath(data):
  mydir, ext = get_rel_idpath(data[DATE_L], data[MSG_ID_L])
  return config.drafts_dir + mydir, ext + ".draft"


def get_rawfilepath(data):
  mydir, ext = get_rel_idpath(data[DATE_L], data[MSG_ID_L])
  return config.old_rawmail_dir + mydir, ext + ".mime"
  

def do_get_referencelist(my_id, refstr):
  reflist = []
  for r in refstr.split(", "):
    rs = r.strip()
    if len(rs) > 0 and rs != my_id:
      reflist.append(rs)
  return reflist


def do_get_parent_id(my_id, reflist):
  if len(reflist) == 0 or reflist[-1] == my_id:
   return None
  return reflist[-1]


def data_to_str(data):
  a = []
  for i,field in enumerate(fields[:-1]):
    a.append(field)
    a.append(": ")
    a.append(data[i])
    a.append("\n")
  a.append(sep)
  a.append(data[BODY_L])
  return "".join(a)


def file_to_data(filename):
  data = []
  try:
    with open(filename) as f:
      for i in range(len(fields)-1):
        line = f.readline()
        data.append(line[len(fields[i]) + 2 : -1])  # erase "field: " and trailing "\n"
      f.readline()  # separator
      data.append(f.read())
  except:
    with open(config.error_log_file, "a") as f:
      f.write("mailfile: could not find " + str(filename) + "\n")
    return list(error_data)
  return data


def write_to_file(filename, data):
  if "/" in filename:  # hopefully!?!
    os.makedirs(filename[:filename.rfind("/")], exist_ok=True)
  with open(filename, "w") as f:
    f.write(data_to_str(data))


# lazily loaded mailfiles
class MailMgr:
  def __init__(self):
    self.ids = msgids.Ids()
    self.datas = {}   # None or not present if unloaded
    
  def ensure_loaded(self, filename):
    if filename not in self.datas or self.datas[filename] is None:
      self.datas[filename] = file_to_data(filename)

  def get(self, filename, ind):
    self.ensure_loaded(filename)
    return self.datas[filename][ind]

  def get_all(self, filename):
    self.ensure_loaded(filename)
    return self.datas[filename]

  def set(self, filename, ind, val):
    self.ensure_loaded(filename)
    self.datas[filename][ind] = val
    write_to_file(filename, self.datas[filename])

  def get_references(self, filename):
    my_id = self.get(filename, MSG_ID_L).strip()
    refstr = self.get(filename, REFS_L)
    return do_get_referencelist(my_id, refstr)

  def get_refby(self, filename):
    s = self.get(filename, REF_BY_L)
    return s.split(", ") if len(s) > 0 else []

  def add_refby(self, filename, msg_id):
    prev_refbys = self.get_refby(filename)
    prev_refbys.append(msg_id)
    self.set(filename, REF_BY_L, ", ".join(prev_refbys))

  # get the mailfile of the message that 'filename' replies to, or None
  def get_parent_id(self, filename):
    return do_get_parent_id(self.get(filename, MSG_ID_L), self.get_references(filename))

  def get_parent(self, filename, ids_to_filenames):
    ref = self.get_parent_id(filename)
    return ids_to_filenames[ref] if ref in ids_to_filenames else None

  def get_tags(self, filename, include_folders=False):
    return tags.get_taglist_from_str(self.get(filename, TAGS_L), include_folders=include_folders)

  def get_tagset(self, filename, include_folders=False):
    return tags.get_tagset_from_str(self.get(filename, TAGS_L), include_folders=include_folders)

  def add_tags(self, filelist, my_taglist):
    taglist = tags.get_taglist_from_list(my_taglist)  # clean, sort tags
    if len(taglist) > 0:
      for filename in filelist:
        new_taglist = tags.join(taglist, self.get_tags(filename))
        self.set(filename, TAGS_L, tags.get_str_from_clean_taglist(new_taglist))

  def remove_tags(self, filelist, my_taglist):
    taglist = tags.get_taglist_from_list(my_taglist)  # clean, sort tags
    if len(taglist) > 0:
      for filename in filelist:
        new_taglist = tags.get_nonmatching(self.get_tags(filename), taglist)
        self.set(filename, TAGS_L, tags.get_str_from_clean_taglist(new_taglist))

  def set_tags(self, filelist, taglist):
    tagstr = tags.get_str_from_taglist(taglist)  # clean tags
    for filename in filelist:
      self.set(filename, TAGS_L, tagstr)

  # if there was an update to these files (added, changed, or deleted),
  # just remove them and re-load at next request
  def updated_filelist(self, filelist):
    for f in filelist:
      if f in self.datas:
        del self.datas[f]

  def create(self, filelist, datalist):
    for f,d in zip(filelist, datalist):
      self.datas[f] = d
    self.ids.add([d[MSG_ID_L] for d in datalist], filelist)
    for f,d in zip(filelist, datalist):
      p_id = do_get_parent_id(d[MSG_ID_L], do_get_referencelist(d[MSG_ID_L], d[REFS_L]))
      if p_id is not None and p_id in self.ids.ids_to_filenames:
        self.add_refby(self.ids.get(p_id), d[MSG_ID_L])

  # when files are deleted
  def trash(self, filelist):
    msg_ids = []
    for f in filelist:
      msg_ids.append(self.get(f, MSG_ID_L))
      if f in self.datas:  # which it should be...
        del self.datas[f]
    self.ids.remove(msg_ids, filelist)


def main_cli():
  if len(sys.argv) <= 1 or "-h" in sys.argv:
    print(usage_str)
    exit(0)
  #filenames = sys.stdin.readlines()
  filenames = [f[:-1] for f in sys.stdin.readlines()]
  mgr = MailMgr()
  if sys.argv[1] == "-g":
    if len(sys.argv) <= 2:
      print(usage_str)
      exit(0)
    if sys.argv[2] == "all":
      print("\n\n".join([data_to_str(mgr.get_all(f)) for f in filenames]))
    else:
      inds = [fields.index(s) for s in sys.argv[2:]]
      print("\n\n".join(["\n".join([mgr.get(f, i) for i in inds]) for f in filenames]))
  elif sys.argv[1] == "-add-tags":
    taglist = sys.argv[2:]
    mgr.add_tags(filenames, taglist)
  elif sys.argv[1] == "-rm-tags":
    taglist = sys.argv[2:]
    mgr.remove_tags(filenames, taglist)
  else:
    print(usage_str)


if __name__ == "__main__":
  main_cli()




####
# bomail.cli.search
#
# Command-line interface to search.
####

import sys
import os, subprocess, shlex
import dateutil.parser
from dateutil import tz
import textwrap

import bomail.config.config as config
import bomail.cli.mailfile as mailfile
import bomail.cli.chstate as chstate
import bomail.util.listmail as listmail
import bomail.util.search_opts as search_opts

my_usage_str = """
Searches through mail for files matching given arguments.

High-level options:
    search.py -h            # print this help
    search.py -num ...      # only print the number of matches
    search.py -sum ...      # print summaries along with filenames
"""
usage_str = my_usage_str + search_opts.options_str

default_max_results = -1  # no max


# given filelist which already matches and is sorted,
# filter 'addlist' and merge into filelist in sorted order
def merge(filelist, addlist, mgr, arg_str):
  # note: shlex.split is quite slow
  query = search_opts.SQ_from_str(arg_str)
  new_addlist = [addlist[i] for i in query.filter(addlist, mgr)]
  filelist += new_addlist  # todo: smarter merge by assuming both are sorted
  filelist.sort(reverse = query.sort_new)


# return list of filenames
def main_argstr(arg_str, mgr, filelist=None):
  # note: shlex.split is quite slow, avoid calling it many times per second
  return main_arglist(shlex.split(arg_str), mgr, filelist=filelist)


# return list of filenames
def main_arglist(args, mgr, filelist=None):
  query = search_opts.SearchQuery(args)
  if filelist is None:
    filelist = listmail.do_list(query, mgr)
  new_filelist = [filelist[i] for i in query.filter(filelist, mgr, listmailed=(filelist is None))]
  new_filelist.sort(reverse = query.sort_new)
  return new_filelist


def main_cli():
  args = sys.argv[1:]
  if "-h" in args:
    print(usage_str)
    exit()
  printout = True
  if "-num" in args:
    printout = False
    i = args.index("-num")
    args = args[:i] + ([] if i+1 >= len(args) else args[i+1:])
  summaries = False
  if "-sum" in args:
    summaries = True
    i = args.index("-sum")
    args = args[:i] + ([] if i+1 >= len(args) else args[i+1:])
  filelist = main_arglist(args, mailfile.MailMgr())
  if printout:
    if summaries:
      mgr = mailfile.MailMgr()
      for fname in filelist:
        print(fname)
        is_sent = bool(mgr.get(fname, mailfile.SENT_L))
        print("    " + mgr.get(fname, mailfile.DATE_L)[:10] + "  " + mgr.get(fname, mailfile.FROM_L if is_sent else mailfile.TO_L)[:64])
        print("    " + mgr.get(fname, mailfile.SUBJ_L)[:76])
        split_lists = [textwrap.wrap(s, 76, replace_whitespace=False) for s in mgr.get(fname, mailfile.BODY_L).split("\n")]
        lines = ["    " + l for splitter in split_lists for l in splitter][:4]
        print("\n".join(lines))
        print()
    else:
      print("\n".join(filelist))
  else:
    print(str(len(filelist)) + " matching messages")


if __name__ == "__main__":
  main_cli()




###
# bomail.guistuff.acts
#
# "acts" are any undoable actions taken by the program,
# such as changing the state of an email, adding tags, or gui actions.
####

import sys

import bomail.config.guiconfig as guicfg
import bomail.config.config as config

import bomail.cli.chstate as chstate
import bomail.cli.mailfile as mailfile
import bomail.guistuff.display as display
import bomail.guistuff.display_fmt as display_fmt
import bomail.guistuff.tabnothread as tabnothread
import bomail.guistuff.tabthread as tabthread


class Acts:
  def __init__(self, gui):
    self.acts = []
    self.act_ind = -1
    self.gui = gui

  def do(self, act):
    self.acts = self.acts[:self.act_ind + 1]  # cut off any old branches
    self.act_ind = len(self.acts) - 1         # just in case the index got too large
    self.acts.append(act)
    return self.redo()

  # return pair a, note
  # where a = kind of redraw to do
  def redo(self):
    self.act_ind += 1
    act = self.acts[self.act_ind]
    return self.exec_act(act)
    
  def exec_act(self, act):
    with open(config.acts_log_file, "a") as f:
      f.write("DO: " + str(act) + "\n")
    if act[0] in ["trash", "open", "closed", "scheduled", "add tags", "remove tags", "add attach", "remove attach"]:
      if len(act[1]) >= 100:  # number 100 chosen via extensive UI studies
        display.redraw_note(self.gui, "Processing command (please wait)...")

    if act[0] == "trash":  # ("trash", filelist, datalist)
      shorts = ", ".join([display_fmt.get_shortened(d, self.gui.mail_mgr) for d in act[1]])
      disp_infos = [t.get_old_display_info() for t in self.gui.tabs]
      chstate.trash(act[1], self.gui.mail_mgr)
      self.gui.update_for_trash(act[1], act[2], disp_infos)
      return "all", "Trashed " + shorts

    elif act[0] in ["open", "closed", "scheduled"]:  # (act, filelist, prevstatelist [, sched])
      shorts = ", ".join([display_fmt.get_shortened(d, self.gui.mail_mgr) for d in act[1]])
      disp_infos = [t.get_old_display_info() for t in self.gui.tabs]
      if act[0] == "open":
        chstate.make_open(act[1], self.gui.mail_mgr)
        note = "Opened " + shorts
      elif act[0] == "closed":
        chstate.make_closed(act[1], self.gui.mail_mgr)
        note = "Closed " + shorts
      else:
        chstate.schedule(act[1], act[3], self.gui.mail_mgr)
        note = "Scheduled for " + act[3].isoformat().replace("T"," ") + ": " + shorts
      self.gui.update_for_change(act[1], disp_infos)
      return "all", note

    elif act[0] == "add tags":  # ("add tags", filelist, addedtagslist, oldtagslists, brandnewtags)
      shorts = ", ".join([display_fmt.get_shortened(d, self.gui.mail_mgr) for d in act[1]])
      disp_infos = [t.get_old_display_info() for t in self.gui.tabs]
      self.gui.mail_mgr.add_tags(act[1], act[2])
      if len(act[4]) > 0:
        self.gui.tag_mgr.add_tags(act[4])
      self.gui.update_for_change(act[1], disp_infos)
      return "all", "Added tags: " + ", ".join(act[2]) + " to " + shorts

    elif act[0] == "remove tags":  # ("remove tags", filelist, removedtagslist, oldtagslists)
      disp_infos = [t.get_old_display_info() for t in self.gui.tabs]
      shorts = ", ".join([display_fmt.get_shortened(d, self.gui.mail_mgr) for d in act[1]])
      self.gui.mail_mgr.remove_tags(act[1], act[2])
      self.gui.update_for_change(act[1], disp_infos)
      return "all", "Removed tags: " + ", ".join(act[2]) + " from " + shorts

    elif act[0] == "add attach":  # ("add attach", filelist, old_attachlist, new_attachlist)
      shorts = ", ".join([display_fmt.get_shortened(d, self.gui.mail_mgr) for d in act[1]])
      disp_infos = [t.get_old_display_info() for t in self.gui.tabs]
      for f in act[1]:
        self.gui.mail_mgr.set(f, mailfile.ATTACH_L, ", ".join([fname.strip() for fname in act[2] + act[3]]))
      self.gui.update_for_change(act[1], disp_infos)
      return "tab", "Added attachment " + ", ".join(act[3]) + " to " + shorts

    elif act[0] == "remove attach":  # ("remove attach", filelist, prev_attachlist, remove_attachlist)
      shorts = ", ".join([display_fmt.get_shortened(d, self.gui.mail_mgr) for d in act[1]])
      disp_infos = [t.get_old_display_info() for t in self.gui.tabs]
      for f in act[1]:
        self.gui.mail_mgr.set(f, mailfile.ATTACH_L, ", ".join([fname.strip() for fname in act[2] if fname.strip() not in act[3]]))
      self.gui.update_for_change(act[1], disp_infos)
      return "tab", "Removed attachment " + ", ".join(act[3]) + " to " + shorts

    elif act[0] == "edit tab":  # ("edit tab", tab_ind, new_str, old_str)
      self.gui.tabs[act[1]].search_str = act[2]
      self.gui.tabs[act[1]].load()
      self.gui.rewrite_tab_searches()
      return "all", "Changed tab search string to: " + act[2]

    elif act[0] == "add tab":  # ("add tab", tab_ind, new_str)
      if guicfg.threads_on:
        self.gui.tabs.insert(act[1], tabthread.TabThread(act[2], self.gui))
      else:
        self.gui.tabs.insert(act[1], tabnothread.TabNoThread(act[2], self.gui))
      self.gui.change_tab_ind(act[1])
      self.gui.rewrite_tab_searches()
      return "all", "Added new tab with search string: " + act[2]

    elif act[0] == "remove tab":  # ("remove tab", tab_ind, tab object, disp_info)
      del self.gui.tabs[act[1]]
      self.gui.change_tab_ind(min(act[1], len(self.gui.tabs)-1))
      self.gui.rewrite_tab_searches()
      return "all", "Removed tab with search string: " + act[2].search_str

    elif act[0] == "move tab":  # ("move tab", tab_ind, amount)
      src = act[1]
      dest = src + act[2]
      self.gui.tabs[src], self.gui.tabs[dest] = self.gui.tabs[dest], self.gui.tabs[src]
      self.gui.change_tab_ind(dest)
      self.gui.rewrite_tab_searches()
      return "all", "Shifted tab " + ("left" if dest < src else "right")

    elif act[0] == "edit file":  # ("edit file", filename, mailfile_line, new_str, old_str)
      self.gui.mail_mgr.set(act[1], act[2], act[3])
      return "all", "Edited [" + mailfile.fields[act[2]] + "] of " + display_fmt.get_shortened(act[1], self.gui.mail_mgr) + " to: " + act[3]

    else:
      return "all", "bomail error: unknown action: " + act[0]


  def undo(self):
    act = self.acts[self.act_ind]
    self.act_ind -= 1
    with open(config.acts_log_file, "a") as f:
      f.write("UNDO: " + str(act) + "\n")

    if act[0] in ["trash", "open", "closed", "scheduled"]:
      if act[0] == "trash":  # ("trash", filelist, datalist)
        chstate.untrash(act[1], act[2], self.gui.mail_mgr)
        prevstates = [d[mailfile.STATE_L] for d in act[2]]
      else:  # (act, filelist, prevstatelist [, sched])
        prevstates = act[2]
      shorts = ", ".join([display_fmt.get_shortened(d, self.gui.mail_mgr) for d in act[1]])
      
      openlist = []
      closedlist = []
      disp_infos = [t.get_old_display_info() for t in self.gui.tabs]
      for f,s in zip(act[1], prevstates):
        if s == "open":
          openlist.append(f)
        elif s == "closed":
          closedlist.append(f)
        else:
          chstate.schedule([f], dateutil.parser.parse(s.split()[1]), self.gui.mail_mgr)
      if len(openlist) > 0:
        chstate.make_open(openlist, self.gui.mail_mgr)
      if len(closedlist) > 0:
        chstate.make_closed(closedlist, self.gui.mail_mgr)
      if act[0] == "trash":
        self.gui.update_for_add(act[1], disp_infos)
      else:
        self.gui.update_for_change(act[1], disp_infos)
      return "all", "Undid change of state to " + act[0] + " for " + shorts

    elif act[0] in ["add tags", "remove tags"]:  # (act, filelist, tagslist, oldtagslists [, brand_new_tags])
      shorts = ", ".join([display_fmt.get_shortened(d, self.gui.mail_mgr) for d in act[1]])
      disp_infos = [t.get_old_display_info() for t in self.gui.tabs]
      for f,prev in zip(act[1], act[3]):
        self.gui.mail_mgr.set_tags([f], prev)
      if act[0] == "add tags" and len(act[4]) > 0:
        self.gui.tag_mgr.remove_tags(act[4])
      self.gui.update_for_change(act[1], disp_infos)
      return "all", "Undid edit of tags for " + shorts

    elif act[0] == "add attach":  # ("add attach", filelist, old_attachlist, new_attachlist)
      shorts = ", ".join([display_fmt.get_shortened(d, self.gui.mail_mgr) for d in act[1]])
      disp_infos = [t.get_old_display_info() for t in self.gui.tabs]
      for f in act[1]:
        self.gui.mail_mgr.set(f, mailfile.ATTACH_L, ", ".join([fname.strip() for fname in act[2]]))
      self.gui.update_for_change(act[1], disp_infos)
      return "tab", "Undid new attachment to " + shorts

    elif act[0] == "remove attach":  # ("remove attach", filelist, prev_attachlist, remove_attachlist)
      shorts = ", ".join([display_fmt.get_shortened(d, self.gui.mail_mgr) for d in act[1]])
      disp_infos = [t.get_old_display_info() for t in self.gui.tabs]
      for f in act[1]:
        self.gui.mail_mgr.set(f, mailfile.ATTACH_L, ", ".join([fname.strip() for fname in act[2]]))
      self.gui.update_for_change(act[1], disp_infos)
      return "tab", "Undid removal of attachment from " + shorts

    elif act[0] == "edit file":  # ("edit file", filename, mailfile_line, new_str, old_str)
      self.gui.mail_mgr.set(act[1], act[2], act[4])
      return "all", "Undid edit of [" + mailfile.fields[act[2]] + "] for " + display_fmt.get_shortened(act[1], self.gui.mail_mgr)

    elif act[0] == "edit tab":  # ("edit tab", tab_ind, new_str, old_str)
      self.gui.tabs[act[1]].search_str = act[3]
      self.gui.tabs[act[1]].is_loaded = False
      self.gui.rewrite_tab_searches()
      return "all", "Undid change of tab search string"

    elif act[0] == "add tab":  # ("add tab", tab_ind, new_str)
      del self.gui.tabs[act[1]]
      self.gui.change_tab_ind(max(0, act[1]-1))
      self.gui.rewrite_tab_searches()
      return "all", "Undid addition of new tab"

    elif act[0] == "remove tab":  # ("remove tab", tab_ind, tab_object, disp_info)
      self.gui.tabs.insert(act[1], act[2])
      act[2].recheck(act[3])  # for changes since it was last used/updated
      self.gui.change_tab_ind(act[1])
      self.gui.rewrite_tab_searches()
      return "all", "Undid removal of tab"

    elif act[0] == "move tab":  # ("move tab", tab_ind, amount)
      dest = act[1]
      src = dest + act[2]
      self.gui.tabs[src], self.gui.tabs[dest] = self.gui.tabs[dest], self.gui.tabs[src]
      self.gui.change_tab_ind(dest)
      self.gui.rewrite_tab_searches()
      return "all", "Undid shift of tab"




####
# bomail.guistuff.display_fmt
#
# Format messages or threads for display.
####

import sys
import time
import curses
import curses.textpad
import re
import time
import textwrap, dateutil.parser, time
from dateutil import tz

import bomail.config.config as config
import bomail.config.guiconfig as guicfg
import bomail.cli.mailfile as mailfile
import bomail.util.addr as addr
import bomail.util.tags as tags

ATTACH_STR = " + "
MARKING_STR = " [ ]"

#---------------------------------------------------------------
# utility stuff, getting display-formatted versions of things

def get_disp_author_from_pair(pair):
  return pair[0] if pair[0] != "" else pair[1]
  
def get_disp_author_from_str(s):
  return get_disp_author_from_pair(addr.str_to_pair(s))

def get_disp_author(data, is_draft=False):
  if is_draft:
    to_strs = [data[l] for l in [mailfile.TO_L, mailfile.CC_L, mailfile.BCC_L]]
    tolists = [addr.str_to_pairlist(s) for s in to_strs]
    return "[DRAFT] " + ", ".join([get_disp_author_from_pair(p) for l in tolists for p in l])
  else:
    return get_disp_author_from_str(data[mailfile.FROM_L])

def get_disp_datestr(data):
  return dateutil.parser.parse(data[mailfile.DATE_L]).astimezone(tz.tzlocal()).strftime(guicfg.datefmt)

indic = lambda x: 1 if x else 0


#---------------------------------------------------------------
# formatting messages / threads into displayable chunks

# return a shortened string describing the email
def get_shortened(filename, mgr):
  data = mgr.get_all(filename)
  author = get_disp_author(data)[:guicfg.short_authorlen]
  subject = data[mailfile.SUBJ_L][:guicfg.short_subjectlen]
  return "[" + subject + "] by " + author

def get_shortened_list(filelist, mail_mgr):
  return "(" + str(len(filelist)) + " messages) " + ", ".join([get_shortened(f, mail_mgr) for f in filelist])


### These routines format a message or thread for display
# They return a pair (msg_lines, attr_data)
# where msg_lines is a list of text lines
# and attr_data is a list of quads (y, x, width, attr)


# get lines, attr_data for standard inbox view
# if use_curses is false, don't fill out attr_data
def format_msg_lines(mail_mgr, screen_width, authorstr, subj, tagstr, datestr, has_attach, is_draft, orig_body, thread_num, thread_total, use_curses=True):
  top_line, attr_data = format_msg_top_line(
      screen_width,
      authorstr,
      subj,
      tagstr if guicfg.show_tags and guicfg.inline_tags else None,
      datestr,
      has_attach,
      is_draft,
      thread_num,
      thread_total,
      use_curses=use_curses)
     
  max_lines = guicfg.lines_per_msg
  msg_lines = [top_line]
  if max_lines == 1:
    return msg_lines, attr_data

  tag_lines = 0
  if guicfg.show_tags and not guicfg.inline_tags:
    # tags get their own line
    tag_lines = 1
    s = " "*(guicfg.authorlen + guicfg.authorpad) + tagstr
    s = s[:screen_width]
    msg_lines.append(s)
    if use_curses:
      attr_data.append((1, guicfg.authorlen + guicfg.authorpad, len(tagstr), curses.color_pair(guicfg.TAGS_CLR_PAIR)))

  bodywidth  = screen_width - guicfg.authorlen - guicfg.authorpad - 1
  indent_len = guicfg.authorlen + guicfg.authorpad
  new_lines, new_data = get_indented_bodylines(orig_body, bodywidth, indent_len, guicfg.remove_newlines, len(msg_lines), use_curses=use_curses)
  msg_lines += new_lines
  attr_data += new_data
  
  # trim msg_lines if needed
  if max_lines >= 1:
    stop_at = max(1, max_lines - indic(guicfg.horiz_line) - indic(guicfg.skip_line) - tag_lines)
    if guicfg.pad_msg_lines:
      while len(msg_lines) < stop_at:
        msg_lines.append("")
    msg_lines = msg_lines[:stop_at]

  if guicfg.horiz_line:
    msg_lines.append("-"*screen_width)
  if guicfg.skip_line:
    msg_lines.append(" "*screen_width)

  return msg_lines, attr_data


# get lines for the standard inbox view with threads off
def get_msg_lines_nothread(mail_mgr, filename, screen_width, use_curses=True):
  data = mail_mgr.get_all(filename)
  return format_msg_lines(
      mail_mgr,
      screen_width,
      get_disp_author(data, filename[-5:] == "draft"),
      data[mailfile.SUBJ_L],
      data[mailfile.TAGS_L],
      get_disp_datestr(data),
      len(data[mailfile.ATTACH_L].strip()) > 0,
      filename[-5:] == "draft",
      data[mailfile.BODY_L],
      None,
      None,
      use_curses)

# get lines for the standard inbox view with threads on
# get data for a thread, formatted as a single displayed set of lines
# current approach: display *oldest NEW message* in thread
#   (that is, the *last* one matching the tab search query)
# plus (num_new / total_in_thread)
# plus display multiple authors and ellipsis if appropriate
def get_msg_lines_thread(mail_mgr, thread_view, screen_width):
  fnames = thread_view.all_files
  datas = [mail_mgr.get_all(f) for f in fnames]
  datas.reverse()  # newest to oldest?
  is_draft_list = [f[-5:] == "draft" for f in fnames]
  is_draft = any(is_draft_list)
  has_attach = any([len(d[mailfile.ATTACH_L].strip()) > 0 for d in datas])
  authors = []
  for i,d in enumerate(datas):
    auth = get_disp_author(d, is_draft_list[i])
    if auth not in authors:
      authors.append(auth)
  authorstr = ", ".join(authors)
  subject = mail_mgr.get(thread_view.repr_file, mailfile.SUBJ_L)
  datestr = get_disp_datestr(mail_mgr.get_all(thread_view.repr_file))
  body = mail_mgr.get(thread_view.repr_file, mailfile.BODY_L)
  tagset = set()
  for d in datas:
    tagset.update(tags.get_tagset_from_str(d[mailfile.TAGS_L], include_folders=False))
  tagstr = ", ".join(sorted(list(tagset)))

  return format_msg_lines(
      mail_mgr,
      screen_width,
      authorstr,
      subject,
      tagstr,
      datestr,
      has_attach,
      is_draft,
      body,
      len(thread_view.matching_set),
      len(thread_view.all_files))


# get format data for a single message to be shown in a thread view
# (linear or tree view)
# here orig_data is the data of the root of the message tree
# return pair lines, attribute_data
def get_thread_view_lines(data, orig_data, is_draft, screen_width, is_collapsed, thread_num, thread_total, level):
  tagstr = data[mailfile.TAGS_L]

  # In the future with tree view, this will be  level * guicfg.thread_indent
  indent_amt = 0
  top_line, attr_data = format_msg_top_line(
      screen_width - indent_amt, 
      get_disp_author(data, is_draft),
      data[mailfile.SUBJ_L],
      tagstr if guicfg.thread_show_tags and guicfg.thread_inline_tags else None,
      get_disp_datestr(data),
      len(data[mailfile.ATTACH_L].strip()) > 0,
      is_draft,
      thread_num,
      thread_total)
     
  max_lines = guicfg.thread_lines_per_collapsed_msg if is_collapsed else -1
  msg_lines = [" "*indent_amt + str(top_line)]

  tag_lines = 0
  if guicfg.thread_show_tags and not guicfg.thread_inline_tags:
    tag_lines = 1
    s = " "*(guicfg.authorlen + guicfg.authorpad) + tagstr
    s = s[:screen_width]
    msg_lines.append(s)
    attr_data.append((1, guicfg.authorlen + guicfg.authorpad, len(tagstr), curses.color_pair(guicfg.TAGS_CLR_PAIR)))

  bodywidth  = screen_width - guicfg.authorlen - guicfg.authorpad - 1
  indent_len = guicfg.authorlen + guicfg.authorpad + indent_amt
  new_lines, new_data = get_indented_bodylines(data[mailfile.BODY_L], bodywidth, indent_len, guicfg.thread_remove_newlines, len(msg_lines))
  msg_lines += new_lines
  attr_data += new_data
  
  # trim msg_lines if needed
  max_lines = guicfg.thread_lines_per_collapsed_msg if is_collapsed else -1
  if max_lines >= 1:
    stop_at = max(1, max_lines - indic(guicfg.thread_horiz_line) - indic(guicfg.thread_skip_line) - tag_lines)
    if guicfg.pad_msg_lines:
      while len(msg_lines) < stop_at:
        msg_lines.append("")
    msg_lines = msg_lines[:stop_at]

  if guicfg.thread_horiz_line:
    msg_lines.append("-"*screen_width)
  if guicfg.thread_skip_line:
    msg_lines.append(" "*screen_width)

  return msg_lines, attr_data


# remove_newlines_type is "all", "some", or "none"
# return (lines_of_text, attr_data)
def get_indented_bodylines(orig_body, bodywidth, indent_len, remove_newlines_type, y_offset, use_curses=True):
  attr_data = []
  indent = " " * indent_len
  if remove_newlines_type == "all":
    return [indent + l for l in textwrap.wrap(orig_body, bodywidth, replace_whitespace=True)], []

  if remove_newlines_type == "some":
    orig_bodylines = [s for s in orig_body.splitlines() if len(s.strip()) > 0]
  else:  # none
    orig_bodylines = [s for s in orig_body.splitlines()]
  wrapped_lines_lst = [textwrap.wrap(line, bodywidth) for line in orig_bodylines]
  y_coord = y_offset
  lines = [indent + l for linelst in wrapped_lines_lst for l in linelst]
  if use_curses:
    for ind,line in enumerate(lines):
      attr_data.append((y_coord+ind, indent_len, bodywidth, curses.color_pair(guicfg.DEFAULT_CLR_PAIR)))
  for orig_ind,line in enumerate(orig_bodylines):
    if len(line) > 0 and line[0] == ">":
      for i in range(len(wrapped_lines_lst[orig_ind])):
        if use_curses:
          attr_data.append((y_coord+i, indent_len, bodywidth, guicfg.quote_attr))
    y_coord += len(wrapped_lines_lst[orig_ind])
  return lines, attr_data


# format given data into a line
# return msg_line (a string) and attr_data (a list of quads (y, x, width, attr))
# tagstr: None if no tags, else a string of tags
# is_draft: Bool
# if thread_num and thread_total are None, ignore
# else display [thread_num of thread_total]
def format_msg_top_line(screen_width, authorstr, subj, tagstr, datestr, has_attach, is_draft, thread_num, thread_total, use_curses=True):
  attr_data = []
  line_lst = []  # join these strings together at the end
  line_len = 0

  # AUTHOR and attachment
  author = authorstr[:guicfg.authorlen]
  line_len = guicfg.authorlen + guicfg.authorpad
  if has_attach:
    line_lst.append(author[:line_len - len(ATTACH_STR)])
    line_lst.append(" "*(line_len - len(ATTACH_STR) - len(author)))
    line_lst.append(ATTACH_STR)
  else:
    line_lst.append(author)
    line_lst.append(" "*(line_len - len(author)))

  if use_curses:
    attr = curses.color_pair(guicfg.AUTHOR_CLR_PAIR) | guicfg.author_attr
    attr_data.append((0, 0, line_len, attr))

  # SUBJECT [X of Y]
  pre_date_width = screen_width - guicfg.datelen - guicfg.datepad - len(MARKING_STR)
  subject = re.sub(r'\s+', " ", subj)
  if thread_total is None:
    subject = subject[:pre_date_width - line_len]
    line_lst.append(subject)
    line_len += len(subject)
    if use_curses:
      attr_data.append((0, guicfg.authorlen + guicfg.authorpad, len(subject), attr))
  else:
    countstr_list = [" ["]
    if thread_num is not None:
      countstr_list.append(str(thread_num))
      countstr_list.append(" of ")
    countstr_list.append(str(thread_total))
    countstr_list.append("]")
    countstr_len = sum(map(len,countstr_list))
    subject = subject[:pre_date_width - line_len - countstr_len]
    line_lst.append(subject)
    line_len += len(subject)
    if use_curses:
      attr_data.append((0, guicfg.authorlen + guicfg.authorpad, len(subject)+1, attr))
    for cs in countstr_list:
      line_lst.append(cs)
    line_len += countstr_len
    if use_curses:
      attr_data.append((0, line_len - countstr_len + 1, countstr_len - 1, curses.color_pair(guicfg.THREAD_CLR_PAIR) | guicfg.author_attr))
  
  # TAGS
  if tagstr is not None:
    # right-align tags, so we get:
    #  ---spaces---tag1, tag2  DATE
    tag_pad = 2
    min_tag_len = 3
    min_tagstart = line_len + tag_pad
    tagspace_avail = pre_date_width - min_tagstart
    if tagspace_avail >= min_tag_len:  # else give up on tags
      tagstr = tagstr[:tagspace_avail]
      tagstart = pre_date_width - len(tagstr)
      num_spaces = tagstart - line_len
      line_lst.append(" "*num_spaces)
      line_lst.append(tagstr)
      if use_curses:
        attr_data.append((0, line_len, num_spaces, attr))
      line_len += num_spaces
      if use_curses:
        attr_data.append((0, line_len, len(tagstr), curses.color_pair(guicfg.TAGS_CLR_PAIR)))
      line_len += len(tagstr)

  # DATE
  num_spaces = screen_width - guicfg.datelen - line_len - len(MARKING_STR)
  if use_curses:
    attr_data.append((0, line_len, num_spaces, attr))
  line_lst.append(" "*num_spaces)
  line_len += num_spaces

  datestr = datestr[:guicfg.datelen]
  if use_curses:
    attr_data.append((0, line_len, len(datestr), attr))
  line_lst.append(datestr)
  line_len += len(datestr)

  # MARKING BOX
  num_spaces = screen_width - len(MARKING_STR) - line_len
  line_lst.append(" "*num_spaces)
  line_lst.append(MARKING_STR)
  if use_curses:
    attr_data.append((0, screen_width - num_spaces - len(MARKING_STR), num_spaces + len(MARKING_STR), attr))

  return ''.join(line_lst), attr_data




####
# bomail.util.addr
#
# Utilities for dealing with email addresses
# and the "address book", which is stored in config.addr_book_file.
####

import sys
import os
from bisect import bisect_left

import bomail.config.config as config
import bomail.cli.mailfile as mailfile
import bomail.util.merge_lines as merge_lines

####
# Python 'types' used:
# pr      A pair of the form ("Person Name", "local@domain")
# prlist  A list of the form [pr, pr, ...]
#
# Plain text stuff:
#
# email_addr  Assumes no commas or quotes!
#             An email address string of the form
#             "Person Name <local@domain>"
#             or "local@domain"
#
# Address book file format: each line is
#    email_addr, N M
# where N and M are integers for num emails sent and received respectively.
####

# Ideally: each email string we can parse should be a recipient from a MIME
# email (RFC5322).
# Currently: each string in raw_str_list should be a comma-separated list
# of recipients each readable by str_to_pair()
# and with the given send and receive counts



# canonical version that ignores case, dots, and +stuff in
# email address
# input is the second part of a pair (e.g. "local@domain")
def canon(s):
  if "@" not in s:
    return s  # no idea what's going on
  local, domain = s.split("@")
  local = local.lower()
  if "+" not in local and "." not in local:
    return s
  if "+" in local:
    local = local[:local.index("+")]
  if "." in local:
    local = local.replace(".","")
  return local + "@" + domain


# GLOBAL VARIABLE
# the user's list of email addresses
my_addr_list = [canon(a) for a in config.my_aliases + [config.email_addr]]


# given a string describing possibly a list of email names and addresses
# take out everything in quotes and replace it with non-quoted names
# e.g. inside namestr, "Stuff Here" and "Here, Stuff" both get changed
# to Stuff Here
def convert_quoted_names(namestr):
  slist = []
  old_quote_ind = -1
  while True:
    new_quote_ind = namestr.find('"', old_quote_ind+1)
    if new_quote_ind == -1:
      slist.append(namestr[old_quote_ind+1:])
      break
    next_ind = namestr.find('"', new_quote_ind+1)
    if next_ind == -1:
      slist.append(namestr[old_quote_ind+1:])
      break
    # everything between last quoted sequence and this one
    slist.append(namestr[old_quote_ind+1: new_quote_ind])
    name = namestr[new_quote_ind+1:next_ind]
    if "," in name:
      namelist = [r.strip() for r in name.split(",")]
      namelist.reverse()
      name = ' '.join(namelist)
    slist.append(name)
    old_quote_ind = next_ind
  return ''.join(slist)


# given plain-text comma-separated list of email addresses,
# return list of pairs
def str_to_pairlist(auth_str):
  if '"' in auth_str:
    auth_str = convert_quoted_names(auth_str)
  str_list = [s.strip() for s in auth_str.split(",") if len(s.strip()) > 0]
  return list(map(str_to_pair, str_list))


# opposite direction
def pairlist_to_str(pair_list):
  return ", ".join([pair_to_str(*r) for r in pair_list])


# return author_name, email_addr
def str_to_pair(auth_str):
  try:  # Author Name <email@addr.domain>
    ind = auth_str.index("<")
    name = auth_str[:ind].strip().strip('"').strip()
    addr = auth_str[ind+1:auth_str.index(">")].strip()
    if "," in name:
      name = " ".join(reversed([s.strip() for s in name.split(",")]))
    return name, addr
  except:  # email@addr.domain
    return "", auth_str.strip()


def pair_to_str(auth, addr):
  if auth == "":
    return addr
  else:
    return auth + " <" + addr + ">"

# given a line from address book,
# return email pair and N, M
def str_to_quad(s):
  if "," not in s:
    p = str_to_pair(s)
    return (p[0], p[1], 0, 0)
  l = s.split(",")
  p = str_to_pair(l[0].strip())
  try:
    counts = l[1].split()
    return (p[0], p[1], int(counts[0].strip()), int(counts[1].strip()))
  except:
    return (p[0], p[1], 0, 0)


# given email pair and two counts,
# return a line for the address book
def quad_to_str(q):
  return pair_to_str(q[0], q[1]) + " , " + str(q[2]) + " " + str(q[3])


# check if the emails go to the same place
def is_recip_eq(p1, p2):
  return canon(p1) == canon(p2)

def is_pair_me(pr):
  return canon(pr[1]) in my_addr_list

def is_str_me(s):
  return is_pair_me(str_to_pair(s))



# Mostly, this class is just used to read the list of addresses
# and to write new ones into the text file
class AddrBook:
  def __init__(self):
    self.pair_to_count = {}    # map (name, email_addr) to (send_count, recv_count)
    self.address_to_pair = {}    # map email_addr to (name, email_addr)
    self.load()


  def load(self):
    if not os.path.exists(config.addr_book_file):
      with open(config.addr_book_file, "w") as f:
        pass
    else:
      with open(config.addr_book_file) as f:
        lines = f.readlines()
      for l in [ll for ll in lines if ll.strip() != ""]:
        q = str_to_quad(l)
        self.pair_to_count[(q[0], q[1])] = (q[2], q[3])
        self.address_to_pair[q[1]] = (q[0], q[1])


  def check_for_new(self, prlist):
    added = False
    for pr in prlist:
      tuple_pr = (pr[0], pr[1])
      if tuple_pr not in self.pair_to_count:
        self.pair_to_count[tuple_pr] = (0, 0)
        self.address_to_pair[pr[1]] = tuple_pr
        added = True
    if added:
      self.rewrite_file()


  # delete all addresses in pairlist
  def remove_addresses(self, pairlist):
    rmlist = [p for p in pairlist if p in self.pair_to_count]
    if len(rmlist) != 0:
      for p in rmlist:
        del self.pair_to_count[p]
        if p[1] in self.address_to_pair:
          del self.address_to_pair[p[1]]
      self.rewrite_file()


  # delete all addresses in pairlist and add their
  # send/receive counts to merge_to
  def merge_addresses(self, merge_to, pairlist):
    c1, c2 = self.pair_to_count[merge_to]
    for pr in pairlist:
      t1, t2 = self.pair_to_count[pr]
      c1 += t1
      c2 += t2
      del self.pair_to_count[pr]
      del self.address_to_pair[pr[1]]
    self.pair_to_count[merge_to] = (c1, c2)
    self.rewrite_file()


  # look up pr in the book
  # if not present, but pr only contains an address
  # that is present, then change to that address
  def lookup_and_change(self, pr):
    if pr in self.pair_to_count:
      return pr
    if pr[0] == "" and pr[1] in self.address_to_pair:
      return self.address_to_pair[pr[1]]
    return pr


  # update our data to include any new addresses and new
  # send/receive counts
  def update_for_new(self, maildatas):
    for data in maildatas:
      fromlist = str_to_pairlist(data[mailfile.FROM_L])
      tolist = str_to_pairlist(data[mailfile.TO_L]) + str_to_pairlist(data[mailfile.CC_L]) + str_to_pairlist(data[mailfile.BCC_L])
      fromlist = [self.lookup_and_change(pr) for pr in fromlist]
      tolist = [self.lookup_and_change(pr) for pr in tolist]
      # increment counts depending on if sent or received
      if data[mailfile.SENT_L] == "True":
        for p in tolist:
          cnt = self.pair_to_count[p] if p in self.pair_to_count else (0, 0)
          self.pair_to_count[p] = (cnt[0] + 1, cnt[1])
        for p in [pr for pr in fromlist if pr not in self.pair_to_count]:
          self.pair_to_count[p] = (0, 0)
          self.address_to_pair[p[1]] = p
      else:
        for p in [pr for pr in tolist if pr not in self.pair_to_count]:
          self.pair_to_count[p] = (0, 0)
          self.address_to_pair[p[1]] = p
        for p in fromlist:
          cnt = self.pair_to_count[p] if p in self.pair_to_count else (0, 0)
          self.pair_to_count[p] = (cnt[0], cnt[1] + 1)
    self.rewrite_file()

 
  # save our address book to the text file
  def rewrite_file(self):
    # add my entry_quads to the dict, then flatten it to a list and write to disk
    entry_quads = [(p[0], p[1], c[0], c[1]) for p,c in self.pair_to_count.items()]
    entry_quads.sort()
    s = "\n".join([quad_to_str(q) for q in entry_quads])
    with open(config.addr_book_file, "w") as f:
      f.write(s)




####
# bomail.util.datestr
#
# Utilities for parsing "date strings"
# which can either be absolute (2050-03-25)
# or relative (p4.5y meaning plus 4.5 years from now)
####

import datetime
import dateutil.parser
from dateutil import tz
from dateutil.relativedelta import relativedelta

datestr_str = """
'date strings' are used in search and scheduling, e.g.
    bomail search -a datestr   # search after given date
    bomail chstate -s datestr  # schedule for given date

datestr can be in:
 (1) Absolute format: yyyy-mm-ddTHH:MM
     or any prefix of this.

 (2) Relative format: p[num][unit]
     meaning plus num of units from now.
     Unit can be y (year), w (week), H (hour), M (minute).
     month is not allowed, to avoid confusing with minute.
     If num has a decimal point, adds the exact amount.
     Else, rounds down to the nearest unit.

 (3) Relative format:  m[num][unit]
     meaning minus num of units before now.
     Same details as (2) apply.

Some examples for absolute date:
  2050              # Jan 1st, 2050 at 00:00
  2050-03-07        # Mar 7th, 2050 at 00:00
  2050-03-07T11     # Mar 7th, 2050 at 11:00

Some examples for relative date:
  p1d               # tomorrow at 00:00, no matter what time it is today
  p1.d              # tomorrow at the same time of day that it is right now
  p1.5d             # 36 hours from now
  m3w               # 3 weeks before now, rounded down to Monday at 00:00
  m3.5w             # exactly 3.5 weeks before now (to the minute)
"""



def get_relativedelta_int(num, suffix):
  if suffix == "y":
    return relativedelta(years=int(num))
  elif suffix == "w":
    return relativedelta(weeks=int(num))
  elif suffix == "d":
    return relativedelta(years=int(num))
  elif suffix == "H":
    return relativedelta(hours=int(num))
  else:  # "M"
    return relativedelta(minutes=int(num))


def get_relativedelta_float(num, suffix):
  if num < 0:
    return -get_relativedelta_float(-num, suffix)
  
  inum = int(num)
  if suffix == "y":
    return relativedelta(years=inum) + get_relativedelta_float((num-inum)*365, "d")
  elif suffix == "w":
    return relativedelta(weeks=inum) + get_relativedelta_float((num-inum)*7, "d")
  elif suffix == "d":
    return relativedelta(days=inum) + get_relativedelta_float((num-inum)*24, "H")
  elif suffix == "H":
    return relativedelta(hours=inum) + get_relativedelta_float((num-inum)*60, "M")
  else:  # "M"
    return relativedelta(minutes=inum)

def get_relativedelta(num, suffix, is_float):
  if is_float:
    return get_relativedelta_float(num, suffix)
  else:
    return get_relativedelta_int(num, suffix)


# convert schedule string to datetime object
def get_datetime(schedstr):
  if schedstr[0] == "p":
    mult = 1.0  # forward in time
  elif schedstr[0] == "m":
    mult = -1.0
  else:
    # absolute
    try:
      return dateutil.parser.parse(schedstr, default=datetime.datetime(1,1,1))
    except:
      return None

  # else: relative
  local_tz = tz.tzlocal()
  now = datetime.datetime.now(local_tz)
  numstr = schedstr[1:-1]
  num = mult*float(numstr)  # forward or back in time
  suffix = schedstr[-1]
  diff = None
  is_float = "." in numstr
  result = now + get_relativedelta(num, suffix, is_float)

  if not is_float:  # round the answer
    if suffix == "y":
      result = datetime.datetime(year=result.year, month=1, day=1)
    elif suffix == "w":
      dayoffset = relativedelta(days=result.weekday())
      result = datetime.datetime(year=result.year, month=result.month, day=result.day) - dayoffset
    elif suffix == "d":
      result = datetime.datetime(year=result.year, month=result.month, day=result.day)
    elif suffix == "H":
      result = datetime.datetime(year=result.year, month=result.month, day=result.day, hour=result.hour)
    elif suffix == "M":
      result = datetime.datetime(year=result.year, month=result.month, day=result.day, hour=result.hour, minute=result.minute)

  return result






####
# bomail.guistuff.tabthread
#
# Tabs when threading is turned on, i.e.
# emails are grouped into conversations.
####

import os
import sys
import subprocess
import curses
import shlex
import functools

import bomail.config.config as config
import bomail.config.guiconfig as guicfg

import bomail.cli.mailfile as mailfile
import bomail.cli.search as search
import bomail.cli.chstate as chstate
import bomail.cli.send as send
import bomail.cli.compose as compose

import bomail.util.addr as addr
import bomail.util.util as util
import bomail.util.thread as thread

import bomail.guistuff.display as display
import bomail.guistuff.display_fmt as display_fmt
import bomail.guistuff.gui_util as gui_util

from bomail.guistuff.tabthreadview import ThreadView


class TabThread:
  def __init__(self, search_str, gui):
    self.search_str = search_str
    self.gui = gui
    self.cursor_ind = 0
    self.display_ind = 0
    self.num_marked = 0
    self.num_displayed = 0

    # list of lists, each of the form:
    #   [threadview_obj, is_marked, disp_data]
    #   where disp_data is either None or (y_coord, lines_to_display, attr_data)
    self.file_data = []

    # list of filenames in this tab
    self.filenames = []
    self.fileset = set()

    # for sorting filenames and tuples
    self.sort_new = not "-sortold" in shlex.split(self.search_str)
    self.file_key = lambda a: self.gui.mail_mgr.get(a, mailfile.DATE_L)
    self.tup_key  = lambda t: self.file_key(self.__get_repres_file_tup(t))
   
    self.is_loaded = False


  # get all the filenames matching our search_str
  def load(self, draw_loading=True, filelist=None):
    self.mode = "thread list"  # or "one thread"
    self.cursor_ind = 0
    self.display_ind = 0
    self.num_marked = 0
    self.num_displayed = 0
    if draw_loading:
      display.draw_loading_screen(self.gui)
    if filelist is None:
      self.filenames = search.main_argstr(self.search_str, self.gui.mail_mgr)
    else:
      self.filenames = filelist
      self.filenames.sort(key=self.file_key, reverse=self.sort_new)
    self.fileset = set(self.filenames)
    thread_data = self.gui.thread_mgr.get_threads_for(self.filenames, self.gui.mail_mgr)
    self.file_data = [[ThreadView(trip, self, self.gui), False, None] for trip in thread_data]
    self.file_data.sort(key=self.tup_key, reverse=self.sort_new)
    self.is_loaded = True


  # get one filename to "represent" a thread
  def __get_repres_file_tup(self, tup):
    return tup[0].repr_file

  def __get_repres_file(self, ind):
    assert(ind >= 0)
    if ind >= len(self.file_data):
      return None
    return self.__get_repres_file_tup(self.file_data[ind])

  
  # Call if we don't know what has changed, but want to try to keep all
  # current settings if possible
  # just consider files currently in our list and/or matching our search string
  def recheck(self, old_disp_info):
    if not self.is_loaded:
      return
    self.__update_all(old_disp_info)
    #display.draw_loading_screen(self.gui)
    #new_filenames = search.main_argstr(self.search_str, self.gui.mail_mgr)
    #new_fileset = set(new_filenames)
    #self.update_for_rm([f for t in self.file_data for f in t[0].all_files if f not in new_fileset], old_disp_info)
    ## assume, worst-case, that all matching files have changed
    #self.update_for_change(new_fileset, old_disp_info)

 
  # lazily load display data and return disp_data, is_unread, is_marked
  # where disp_data = (msg_y, lines, attr_data)
  def get_display_data(self, ind):
    if self.file_data[ind][2] is None:
      height, width = self.gui.tab_area.getmaxyx()
      lines, attr_data = display_fmt.get_msg_lines_thread(self.gui.mail_mgr, self.file_data[ind][0], width)
      self.file_data[ind][2] = [0, lines, attr_data]
    is_unread = any([f in self.gui.unread_set for f in self.file_data[ind][0].all_files])
    is_marked = self.file_data[ind][1]
    return self.file_data[ind][2], is_unread, is_marked


  # get info about current cursor and scrolling settings
  # return tuple (filenames_under_cursor, sort_key_of_cursor, display_index)
  def get_old_display_info(self):
    if not self.is_loaded:
      return (None, None, 0)
    elif len(self.file_data) == 0:
      return (None, None, 0)
    old_cursor_tup = self.file_data[self.cursor_ind]
    old_cursor_fnames = set(old_cursor_tup[0].all_files)
    old_cursor_key = self.tup_key(old_cursor_tup)
    old_display_ind = self.display_ind
    return (old_cursor_fnames, old_cursor_key, old_display_ind)


  # given the same tuple returned by get_old_display_info
  # but now that files are updated,
  # attempt to set the display similar to how it was
  def __update_display(self, disp_info):
    self.cursor_ind, self.display_ind = 0, 0
    if len(self.file_data) == 0:
      return

    old_cursor_fnames, old_cursor_key, old_display_ind = disp_info
    if old_cursor_key is None:
      return
    self.display_ind = old_display_ind

    found_thread = False
    for ind,tup in enumerate(self.file_data):
      for new_f in tup[0].all_files:
        if new_f in old_cursor_fnames:
          self.cursor_ind = ind
          found_thread = True
          break
    if not found_thread:
      self.cursor_ind = util.bisect_left_key(self.file_data, old_cursor_key, key=self.tup_key, reverse=self.sort_new)
      if self.cursor_ind >= len(self.file_data):
        self.cursor_ind = len(self.file_data) - 1
    display.redraw_tab_msg(self, self.gui, refresh_display=False)  # set msg_ys
    self.__scroll_up(amt=0)
    self.__scroll_down(amt=0)


  # save some info that we'll use to reload
  def __save_file_stuff(self):
    old_mode = self.mode
    old_disp_file = self.__get_repres_file(self.cursor_ind)
    oldfile_to_threaddata = {}
    for thread_data, marked, disp_data in self.file_data:
      for f in thread_data.matching_set:
        oldfile_to_threaddata[f] = thread_data
    num_old_files = len(self.file_data)
    return (old_mode, old_disp_file, oldfile_to_threaddata, num_old_files)


  # use old info to keep as much the same as possible
  # after an update to our files
  def __restore_file_stuff(self, old_tup, disp_info):
    old_mode, old_disp_file, oldfile_to_threaddata, num_old_files = old_tup
    # first, thread objects
    for i,tup in enumerate(self.file_data):
      is_new = True
      for f in tup[0].matching_set:
        if f in oldfile_to_threaddata:
          is_changed = tup[0].reuse_disp_data(oldfile_to_threaddata[f])
          if not is_changed:
            is_new = False
          break
    # second, cursor and display index
    self.__update_display(disp_info)
    # third, mode
    if old_mode == "one thread":
      new_display_file = self.__get_repres_file(self.cursor_ind)
      if old_disp_file is not None and new_display_file is not None and new_display_file == old_disp_file:
        self.mode == "one thread"


  # note our old filenames may not exist anymore!
  def __update_all(self, disp_info):
    if not self.is_loaded:
      return
    # load from scratch
    old_tup = self.__save_file_stuff()
    num_old_files = old_tup[-1]
    self.load(draw_loading = (num_old_files >= 200))  # 200 chosen via extensive user studies
    self.__restore_file_stuff(old_tup, disp_info)
        

  # update for the case these files may have changed (but still exist)
  # a more general scenario than update_for_add
  def update_for_change(self, change_list, disp_info):
    if not self.is_loaded:
      return
    match_list = search.main_argstr(self.search_str, self.gui.mail_mgr, filelist=change_list)
    add_list = []
    ch_list = []
    for f in match_list:
      if f in self.fileset:  # files previously present, changed, still present
        ch_list.append(f)
      else:  # files previously not present, changed, present
        add_list.append(f)
    # files previously present, changed, no longer present
    match_set = set(match_list)
    rm_set = set([f for f in change_list if f in self.fileset and f not in match_set])
    updated_list = [f for f in self.filenames if f not in rm_set] + add_list

    with open(config.error_log_file, "a") as f:
      f.write("Updating for change:\nCh list: " + str(ch_list) + "\nMatch list: " + str(match_list) + "\nRm set: " + str(rm_set) + "\n\n")

    old_tup = self.__save_file_stuff()
    draw_load = abs(len(updated_list) - len(self.filenames)) >= 100  # 100 chosen via UI study
    self.load(draw_loading=draw_load, filelist=updated_list)
    self.__restore_file_stuff(old_tup, disp_info)
    

  # update for new files added to system
  # they must not already be in the tab!
  def update_for_add(self, add_list, disp_info):
    if not self.is_loaded:
      return
    add_filenames = search.main_argstr(self.search_str, self.gui.mail_mgr, filelist=add_list)
    if len(add_filenames) == 0:
      return
    old_tup = self.__save_file_stuff()
    self.load(draw_loading=(len(add_filenames)>=200), filelist=self.filenames + add_filenames)
    self.__restore_file_stuff(old_tup, disp_info)
    

  # these files no longer match our search string
  # or are deleted if is_trash
  def update_for_rm(self, rm_list, disp_info, is_trash=False):
    if not self.is_loaded:
      return
    rm_set = set(rm_list)
    new_filenames = [f for f in self.filenames if f not in rm_set]
    if len(new_filenames) == len(self.filenames):
      return  # no change
    old_tup = self.__save_file_stuff()
    self.load(draw_loading=(len(new_filenames)>=200), filelist=new_filenames)
    self.__restore_file_stuff(old_tup, disp_info)


  def update_for_trash(self, trash_list, data_list, disp_info):
    self.update_for_rm(trash_list, disp_info, is_trash=True)


  # return all filenames to be affected by the current action
  def get_curr_filenames(self):
    if not self.is_loaded:
      self.load()
    if len(self.file_data) == 0:
      return []
    elif self.mode == "one thread":
      return self.file_data[self.cursor_ind][0].get_curr_filenames()
    elif self.num_marked == 0:  # cursor thread's files only
      return self.file_data[self.cursor_ind][0].all_files
    else:  # all marked threads' files
      return [f for t in self.file_data if t[1] for f in t[0].all_files]


  def __scroll_up(self, amt=1):
    self.cursor_ind -= amt
    if self.cursor_ind < 0:
      self.cursor_ind = 0
    if self.display_ind > self.cursor_ind:
      self.display_ind = self.cursor_ind

  def __scroll_down(self, amt=1):
    self.cursor_ind += amt
    if self.cursor_ind >= len(self.file_data):
      self.cursor_ind = max(0, len(self.file_data) - 1)
    if self.cursor_ind >= self.display_ind + self.num_displayed:
      self.display_ind = self.cursor_ind

  def __scroll_down_page(self):
    self.__scroll_down(self.display_ind + self.num_displayed - self.cursor_ind)

  # attempt to scroll up until the previous cursor message leaves the screen
  def __scroll_up_page(self):
    old_cursor_ind = self.cursor_ind
    while self.cursor_ind > 0:
      self.cursor_ind -= 1
      display.redraw_tab_msg(self, self.gui, refresh_display=False)  # set msg_ys
      if self.cursor_ind + self.num_displayed <= old_cursor_ind:
        break
    self.display_ind = self.cursor_ind


  # return mode, note
  def process_key(self, key):
    if not self.is_loaded:
      self.load()

    mode, note = "all", ""
    curr_filelist = self.get_curr_filenames()
    
    # if no files, only allowable action is new draft
    if len(curr_filelist) == 0:
      if key == guicfg.WRITE_KEY:
        return gui_util.go_compose_draft(self.gui, "n", None)  # new blank draft
      return "note", "Key not recognized / nothing to do"

    # Now assume len(curr_filelist) > 0

    # changing state
    if key == guicfg.OPEN_KEY:
      mode, note = self.gui.acts.do(("open", curr_filelist, [self.gui.mail_mgr.get(f, mailfile.STATE_L) for f in curr_filelist]))
    elif key == guicfg.CLOSE_KEY:
      mode, note = self.gui.acts.do(("closed", curr_filelist, [self.gui.mail_mgr.get(f, mailfile.STATE_L) for f in curr_filelist]))
    elif key == guicfg.SCHEDULE_KEY:
      mode, note = gui_util.go_schedule(self.gui, curr_filelist)
    elif key == guicfg.TRASH_KEY:
      mode, note = self.gui.acts.do(("trash", curr_filelist, [self.gui.mail_mgr.get_all(f) for f in curr_filelist]))

    # tags
    elif key == guicfg.ADD_TAGS_KEY:
      mode, note = gui_util.go_add_tags(self.gui, curr_filelist)
    elif key == guicfg.REMOVE_TAGS_KEY:
      mode, note = gui_util.go_remove_tags(self.gui, curr_filelist)


    # Other things are treated differently if mode is one thread
    elif self.mode == "one thread":
      return self.file_data[self.cursor_ind][0].process_key(key)


    # navgation type keys
    elif key == guicfg.DOWN_KEY:
      self.__scroll_down()
      mode, note = "tab", ""
    elif key == guicfg.UP_KEY:
      self.__scroll_up()
      mode, note = "tab", ""
    elif key == guicfg.WAY_DOWN_KEY:
      self.__scroll_down_page()
      mode, note = "tab", ""
    elif key == guicfg.WAY_UP_KEY:
      self.__scroll_up_page()
      mode, note = "tab", ""

    # mark/unmark messages
    elif key == guicfg.MARK_KEY:
      name = display_fmt.get_shortened(self.file_data[self.cursor_ind][0].repr_file, self.gui.mail_mgr)
      if self.file_data[self.cursor_ind][1]:  # is marked
        self.file_data[self.cursor_ind][1] = False
        self.num_marked -= 1
        mode, note = "tab", "Un-marked " + name
      else:
        self.file_data[self.cursor_ind][1] = True
        self.num_marked += 1
      mode, note = "tab", "Marked " + name

    elif key == guicfg.MARK_ALL_KEY:
      if self.num_marked == len(self.file_data):  # unmark all
        for i in range(len(self.file_data)):
          self.file_data[i][1] = False
        self.num_marked = 0
        mode, note = "tab", "Marked none"
      else:
        for i in range(len(self.file_data)):
          self.file_data[i][1] = True
        self.num_marked = len(self.file_data)
        mode, note = "tab", "Marked all"

    # writing and viewing
    elif key == guicfg.WRITE_KEY:
      filename = curr_filelist[0]
      mode, note = gui_util.go_write_key(self.gui, filename)

    elif key == guicfg.VIEW_KEY or (guicfg.arrowkey_nav and key == "KEY_RIGHT") or (not guicfg.arrowkey_nav and key == guicfg.RIGHT_KEY):
      self.mode = "one thread"
      return "all", ""

    # sending
    elif key == guicfg.SEND_KEY:
      filename = curr_filelist[0]
      if len(curr_filelist) == 1 and filename[-5:] == "draft":
        return gui_util.go_send(self.gui, filename)
      else:
        return "note", "Cannot send: not a draft"
    else:
      return "note", "Input key not recognized"
    return mode, note




####
# bomail.util.remove_lines
#
# Removes matching lines from a file.
####

import sys
import os
import tempfile

import bomail.util.util as util

usage_str = """
Usage:
    remove_lines.py -h         # print this help
    remove_lines.py filename

Reads line "s1" from stdin and removes first line in file containing s1.
Then reads line "s2" from stdin and removes next line in file containing s2.
Etc.

So s1, s2, ... should be in the same sorted order as lines in the file.
"""


# remove from filename all lines containing a string in the set s_set
# scheduled list style: check if second part of line is in s_set, if so remove it
def do_sched(list_filename, s_set):
  if not os.path.exists(list_filename):
    return
  made_change = False

  # write to a new temp file "tof"
  with open(list_filename) as fromf, tempfile.NamedTemporaryFile(dir=".", mode="w", delete=False) as tof:
    for temp in fromf:
      pr = temp.split()
      if len(pr) >= 2 and pr[1] in s_set:
        s_set.remove(pr[1])
        made_change = True
        continue
      tof.write(temp)
  if made_change:  # otherwise, no need to rewrite file
    util.mv_file(tof.name, list_filename) 
  else:
    os.remove(tof.name)


def do_sorted(list_filename, s_list):
  if not os.path.exists(list_filename):
    return
  made_change = False
  curr_i = 0

  # write to a new temp file "tof"
  with open(list_filename) as fromf, tempfile.NamedTemporaryFile(dir=".", mode="w", delete=False) as tof:
    for temp in fromf:
      while curr_i < len(s_list) and s_list[curr_i] < temp[:-1]:
        curr_i += 1
      if curr_i < len(s_list) and s_list[curr_i] == temp[:-1]:
        made_change = True
        curr_i += 1 # assume no duplicates
        continue
      tof.write(temp)
  if made_change:  # otherwise, no need to rewrite file
    util.mv_file(tof.name, list_filename) 
  else:
    os.remove(tof.name)


if __name__ == "__main__":
  if len(sys.argv) != 2 or "-h" in sys.argv:
    print(usage_str)
    exit(0)

  # could be more efficient by reading as needed...
  # also remove trailing newlines
  slist = [s[:-1] for s in sys.stdin.readlines()]
  do_sorted(sys.argv[1], sorted(slist))




####
# bomail.util.listmail
#
# (Quickly) list emails matching date range or simple state queries.
# The idea is it's fast because it doesn't have to open the files.
####

import sys
import os.path

import bomail.config.config as config
import bomail.cli.mailfile as mailfile
import bomail.util.search_opts as search_opts

usage_str = """
Prints a list of filenames of emails matching some basic filters.
All arguments are optional.
 -h              print this help

Use either, both or neither:
 -a datestr   after date (e.g. yyyy-mm-dd)
 -b datestr   before date

Use at most one:
 -open           restrict to open
 -scheduled      restrict to scheduled
 -draft          drafts only
 -attach         only those having attachments
"""

# after and before may be none
# if f_out is None, return a list of strings,
# otherwise print them to f_out
def list_all(after, before, f_out, mgr, dirname=config.email_dir):
  if not os.path.exists(dirname):
    return []
  all_years = os.listdir(dirname)
  if len(all_years) == 0:
    return []
  output = []
  for year_str in all_years:
    if after is not None and year_str < after[:4]:   continue
    if before is not None and year_str > before[:4]: continue
    year_prefix = year_str + "-"
    year_dir = dirname + year_str + "/"
    all_days = os.listdir(year_dir)
    for day_str in all_days:
      date_str = year_prefix + day_str
      if after is not None and date_str[:len(after)] < after:   continue
      if before is not None and date_str[:len(before)] > before: continue
      day_dir = year_dir + day_str + "/" 
      for s in os.listdir(day_dir):
        if len(s) >= 5 and (s[-4:] == "mail" or s[-5:] == "draft"):
          if f_out is None:
            output.append(day_dir + s)
          else:
            print(day_dir + s, file=f_out)
  return output

# openlist is sorted by reverse lexicographic order
def list_open(after, before, f_out, mgr):
  if not os.path.exists(config.openlist_file):
    return []
  output = []
  with open(config.openlist_file) as f:
    for line in f:
      filename = line[:-1] if (len(line) > 0 and line[-1] == "\n") else line
      if search_opts.datestr_matches(search_opts.get_date_from_filename(filename), after, before):
        if f_out is None: 
          output.append(filename)
        else:
          print(filename, file=f_out)
  return output

# scheduledlist is sorted by schedule date, not file date!
# have to check date of file!
# lines have the form "scheduledate emailfilename"
def list_scheduled(after, before, f_out, mgr):
  if not os.path.exists(config.scheduledlist_file):
    return []
  output = []
  with open(config.scheduledlist_file) as f:
    for line in f:
      if len(line.strip()) == 0:
        continue
      filename = line.split()[1]
      if filename[-1] == "\n":
        filename = filename[:-1]
      datestr = search_opts.get_date_from_filename(filename)
      if search_opts.datestr_matches(datestr, after, before):
        if f_out is None:
          output.append(filename)
        else:
          print(filename, file=f_out)
  return output

def list_drafts(after, before, f_out, mgr):
  return list_all(after, before, f_out, mgr, config.drafts_dir)

# similar to list_all, but use the attach directories (only the ones that exist)
def list_attachonly(after, before, f_out, mgr):
  if not os.path.exists(config.attach_dir):
    return []
  all_years = os.listdir(config.attach_dir)
  if len(all_years) == 0:
    return []
  output = []
  for year_str in all_years:
    if after is not None and year_str < after[:4]:   continue
    if before is not None and year_str > before[:4]: continue
    year_prefix = year_str + "-"
    year_dir = config.attach_dir + year_str + "/"
    all_days = os.listdir(year_dir)
    for day_str in all_days:
      date_str = year_prefix + day_str
      if after is not None and date_str[:len(after)] < after:   continue
      if before is not None and date_str[:len(before)] > before: continue
      day_dir = year_dir + day_str + "/" 
      for s in os.listdir(day_dir):
        if s[-1] == "/":
          s = s[:-1]
        mailfile_path = config.email_dir + year_str + "/" + day_str + "/" + s
        if f_out is None:
          output.append(mailfile_path)
        else:
          print(mailfile_path, file=f_out)
  return output
   


# query is a search_opts.SearchQuery
def do_list(query, mgr, f_out=None):
  after, before, draft_only, is_open, scheduled, attachonly = query.get_list_opts()

  # pick an order over these options and just use one of them
  if draft_only:
    return list_drafts(after, before, f_out, mgr)
  elif is_open:
    return list_open(after, before, f_out, mgr)
  elif scheduled:
    return list_scheduled(after, before, f_out, mgr)
  elif attachonly:
    return list_attachonly(after, before, f_out, mgr)
  else:
    return list_all(after, before, f_out, mgr)


# f_out is a writable buffer-like object or None
# if None, return a list of filenames
def main(args, mgr, f_out=None):
  query = search_opts.SearchQuery(args)
  return do_list(query, mgr, f_out=f_out)


if __name__ == "__main__":
  if "-h" in sys.argv:
    print(usage_str)
  else:
    main(sys.argv[1:], mailfile.MailMgr(), f_out=sys.stdout)




####
# bomail.guistuff.gui_util
#
# various UI actions, like Send, that get called from several places.
####

import os
import subprocess
import curses

import bomail.config.guiconfig as guicfg

import bomail.cli.compose as compose
import bomail.cli.send as send
import bomail.cli.mailfile as mailfile

import bomail.guistuff.display as display
import bomail.guistuff.display_fmt as display_fmt
import bomail.guistuff.get_txt as get_txt

import bomail.util.addr as addr
import bomail.util.datestr as datestr


# return string message to the user
def go_send(gui, filename):
  attach_str = gui.mail_mgr.get(filename, mailfile.ATTACH_L)
  if len(attach_str) > 0 and not all(os.path.exists(f) and os.path.isfile(f) for f in attach_str.split(", ")):
    return "all", "Error (did not send): could not find an attachment!"
  lines = ["y  yes, send the email",
           "n  no, cancel"]
  display.draw_bottom_text(gui, lines)
  display.redraw_note(gui, "")
  key = gui.screen.getkey()
  if key == "y":
    disp_infos = [t.get_old_display_info() for t in gui.tabs]
    display.redraw_note(gui, "Sending...")
    num_sent, err_str = send.main([filename], gui.mail_mgr)
    if num_sent == 1:
      gui.acts.exec_act(("trash", [filename], [gui.mail_mgr.get_all(filename)]))
      return "all", "Sent mail!"
    else:
      return "all", err_str
  else:
    return "all", "Cancelled, did not send"


# return mode, note
def go_compose_draft(gui, key, filename, quote=True):
  if key == "n" or filename is None:
    recip_pairs = get_txt.get_recips(gui.tab_area, gui.addr_book)

    gui.tab_area.clear()
    gui.tab_area.refresh()
    subj = get_txt.get_subj_line(gui, "")
    if subj is None:
      return "all", "Cancelled"
    new_filename = compose.new_compose(subj, addr.pairlist_to_str(recip_pairs), "", "", "", gui.mail_mgr)
  elif key == "a":
    new_filename = compose.reply_all(filename, gui.mail_mgr, quote)
  elif key == "o":
    new_filename = compose.reply_one(filename, gui.mail_mgr, quote)
  elif key == "r":
    new_filename = compose.reply_all(filename, gui.mail_mgr, quote, include_sender=False)
  elif key == "f":
    recip_pairs = get_txt.get_recips(gui.tab_area, gui.addr_book)
    new_filename = compose.forward(filename, addr.pairlist_to_str(recip_pairs), gui.mail_mgr, quote)
  elif key == "e":
    new_filename = filename
  else:
    return "all", "Cancelled"
  disp_infos = [t.get_old_display_info() for t in gui.tabs]
  subprocess.call(guicfg.edit_prog + " \"" + new_filename + "\"", shell=True)
  gui.mail_mgr.updated_filelist([new_filename])
  curses.curs_set(False)
  gui.update_for_change([new_filename], disp_infos)
  return "all", "Edited " + display_fmt.get_shortened(new_filename, gui.mail_mgr)


# if the WRITE key was pressed, get the kind of draft to create
# given current filename
# return mode, note, new_filename
def go_write_key(gui, filename):
  if filename[-5:] == "draft":
    return go_write_key_draft(gui, filename)
  else:
    return go_write_key_email(gui, filename)


def go_write_key_email(gui, filename):
  lines = [
    "a  reply-all        f  forward",
    "o  reply-one        n  new draft",
    "r  reply-recipient  z  CANCEL"]
  display.draw_bottom_text(gui, lines)
  display.redraw_note(gui, "")
  key = gui.screen.getkey()
  if key in ["a", "o", "r", "f", "n"]:
    return go_compose_draft(gui, key, filename)
  else:
    return "all", "Cancelled"


def go_write_key_draft(gui, filename):
  lines = [
    "n  new draft          t/y  add/remove recip  a/x  add/remove attachment",
    "e  edit this draft    c/v  add/remove CC       z  CANCEL",
    "s  edit subject line  b/m  add/remove BCC"]
  display.draw_bottom_text(gui, lines)
  display.redraw_note(gui, "")
  key = gui.screen.getkey()
  if key in ["n", "e"]:
    return go_compose_draft(gui, key, filename)
  
  elif key == "s":
    s = get_txt.get_subj_line(gui.tab_area, gui.mail_mgr.get(filename, mailfile.SUBJ_L))
    if s is None:
      return "all", "Cancelled"
    gui.mail_mgr.set(filename, mailfile.SUBJ_L, s)
    return "all", "Edited subject line of " + display_fmt.get_shortened(filename, gui.mail_mgr)

  elif key == "a":
    s = get_txt.get_path(gui.tab_area, gui.mail_mgr)
    if s is None:
      return "all", "Cancelled"
    old_attachs = gui.mail_mgr.get(filename, mailfile.ATTACH_L).strip()
    old_attachlist = [] if len(old_attachs) == 0 else [f.strip() for f in old_attachs.split(", ")]
    return gui.acts.do(("add attach", [filename], old_attachlist, [f.strip() for f in s]))

  elif key == "x":
    old_attach_str = gui.mail_mgr.get(filename, mailfile.ATTACH_L).strip()
    if len(old_attach_str) == 0:
      return "all", "No attachments to remove"
    old_attachlist = [f.strip() for f in old_attach_str.split(", ")]
    s = get_txt.get_sublist_from_list(gui.tab_area, old_attachlist, [], allow_new=False)
    if s is None:
      return "all", "Cancelled"
    return gui.acts.do(("remove attach", [filename], old_attachlist, s))

  elif key in ["t", "y"]:
    mail_line = mailfile.TO_L
  elif key in ["c", "v"]:
    mail_line = mailfile.CC_L
  elif key in ["b", "m"]:
    mail_line = mailfile.CC_L
  else:
    return "all", "Cancelled"

  # get the recipients and, if nonempty, do it
  old_recip_line = gui.mail_mgr.get(filename, mail_line)
  old_recip_pairs = addr.str_to_pairlist(old_recip_line)
  if key in ["t", "c", "b"]:
    # get new recips, add to mail_line
    new_recip_pairs = get_txt.get_recips(gui.tab_area, gui.addr_book, already_recip_pairs=old_recip_pairs)
    if len(new_recip_pairs) == 0:
      return "all", "Cancelled"
    new_recip_line = addr.pairlist_to_str(old_recip_pairs + new_recip_pairs)
  else:
    # get from old recips, remove from mail_line
    remove_recip_pairs = get_txt.get_recips(gui.tab_area, gui.addr_book, recip_pairs=old_recip_pairs, allow_new_entries=False)
    if len(remove_recip_pairs) == 0:
      return "all", "Cancelled"
    new_recip_line = addr.pairlist_to_str([r for r in old_recip_pairs if r not in remove_recip_pairs])
  return gui.acts.do(("edit file", filename, mail_line, new_recip_line, old_recip_line))


# return mode, note
def go_schedule(gui, filelist):
  sched_txt = get_txt.get_sched_txt(gui)
  if sched_txt is None:
    return "tab", "Cancelled"
  sched = datestr.get_datetime(sched_txt)
  if sched is None:
    return "tab", "Cancelled: could not parse input"
  return gui.acts.do(("scheduled", filelist, [gui.mail_mgr.get(f, mailfile.STATE_L) for f in filelist], sched))


def go_add_tags(gui, filelist):
  prevtags = [gui.mail_mgr.get_tags(f) for f in filelist]
  oldtaglist = sorted(list(set.intersection(*[gui.mail_mgr.get_tagset(f) for f in filelist])))
  newtags = get_txt.get_tags(gui.tab_area, [t for t in gui.tag_mgr.tags if t not in oldtaglist])
  if newtags is None or len(newtags) == 0:
    return "tab", "Cancelled"
  else:
    alltags = set(gui.tag_mgr.tags)
    unknown = [t for t in newtags if t not in alltags]
    return gui.acts.do(("add tags", filelist, newtags, prevtags, unknown))


def go_remove_tags(gui, filelist):
  prevtags = [gui.mail_mgr.get_tags(f) for f in filelist]
  oldtaglist = sorted(list(set.union(*[gui.mail_mgr.get_tagset(f) for f in filelist])))
  removeme = get_txt.get_tags(gui.tab_area, oldtaglist, allow_new=False)
  with open("temp.txt", "a") as ferr:
    ferr.write("REMOVING THESE TAGS:\n" + str(removeme) + "\n\n")
  if removeme is None or len(removeme) == 0:
    return "tab", "Cancelled"
  else:
    return gui.acts.do(("remove tags", filelist, removeme, prevtags))




####
# bomail.config.config
#
# Main configuration file.
####

import os, sys

bomailrc_filename = ".bomailrc"


#--------------------------------------------------------------
setup_success_str = """
Basic setup successful! Please edit ~/.bomailrc for:
 * SMTP server information, in order to send email
 * location to find new emails (~/mail/new by default)
 * location to put processed emails (~/mail/cur by default)
 * other options.

Run bomail to begin...
"""


#--------------------------------------------------------------
bomail_tabcomplete_str = """

_bomail()
{
    local cur=${COMP_WORDS[COMP_CWORD]}
    COMPREPLY=( $(compgen -W "gui process check_sched search chstate compose mailfile send stats clean_tags" -- $cur) )
}
complete -F _bomail bomail

"""


#--------------------------------------------------------------
sample_mailhandlers_str = """## This file controls handling of new emails in bomail.
## In this file, lines starting with '#' are comments.

## ------------------------------------------------------
## Section 1: Special options
##
## Comment each option out by prepending '#' to turn if off.
## ------------------------------------------------------

## Tag replies to a thread with parent message's tags
-autotagreplies

## Use tags from last line of email if of the form "tags: tag1, ..."
-lastlinetags

## Automatically close any sent mail
-close-sent


## ---------------------------------------------------
## Section 2: Custom handlers
##
## You write these.
## ---------------------------
##
## Format of each handler (blank line between handlers):
##
## search_string
## action_string
##
## search_string1
## OR search_string2
## OR search_string3
## action_string1
## AND action_string2
## AND action_string3
##
## search_string is a command-line argument to search
## actions are (for 'schedstr' see bomail chstate -h):
##   open
##   close
##   schedule schedstr
##   tag tag1, tag2
## ---------------------------------------------------



"""

#--------------------------------------------------------------
sample_config_str = """## bomail configuration file
##
## All lines beginning with # are comments.
## To 'uncomment' a line, delete the initial #.



## ============================================================
## User info: must edit everything in this section!

name = YOUR NAME
email_addr = USER@DOMAIN.COM

#alias_addresses = ALSO_ME@DOMAIN.COM, ALSO_METOO@DOMAIN.COM

smtp_servername = smtp.gmail.com
smtp_port = 587

## If commented, tries to read password from ~/.getmail/getmailrc
#smtp_password = PASSWORD_FOR_email_addr



## ============================================================
## Organization options: should confirm these are okay

bomail_location = ~/bomail
new_rawmail_location = ~/mail/new
processed_rawmail_location = ~/mail/cur



## ============================================================
## Important UI options

## Comment next line to turn off threading (conversations)
threads_on

read_program = less
edit_program = vim

## Comment next line to use vim-style navigation (hjkl for left/down/up/right)
#arrowkey_navigation

## Options are light1, light2, dark1, dark2
colorscheme = dark1



## ============================================================
## Other UI options

## Options are none, some, all
strip_newlines_in_msg_preview = some

## Num lines (including subject/author/date and preview lines) per message
total_lines_per_msg = 5
threadview_total_lines_per_msg = 10

## Number of characters to use ("len") and for spacing ("pad")
date_len = 20
date_pad = 2
author_len = 24
author_pad = 2

## Comment (uncomment) to turn off (on)
#hline_between_msgs
skipline_between_msgs
show_tags
tags_on_topline


"""
#--------------------------------------------------------------


home = os.getenv("HOME")
config_file = os.path.join(home, bomailrc_filename)


# Initial setup!
if not os.path.exists(config_file):
  print("\nWelcome to bomail!")
  sys.stdout.write("Name (as it will appear in From line): ")
  sys.stdout.flush()
  myname = sys.stdin.readline().strip()
  sys.stdout.write("Email address (primary, i.e. send email from): ")
  sys.stdout.flush()
  myaddr = sys.stdin.readline().strip()
  sys.stdout.write("Data files location? (leave blank for default of ~/bomail/): ")
  sys.stdout.flush()
  datadir = sys.stdin.readline().strip()
  if len(datadir) == 0:
    datadir = "~/bomail"
  sys.stdout.write("Add bomail command tab-completion to your ~/.bashrc file? (y/n): ")
  sys.stdout.flush()
  tabcomplete = sys.stdin.readline().strip()

  if len(tabcomplete) > 0 and tabcomplete[0] in ["y", "Y"]:
    with open(os.path.join(home, ".bashrc"), "a") as f:
      f.write(bomail_tabcomplete_str)

  # this is not ideal because it doesn't change if the configuration at
  # the bottom of the file changes
  metadata_dir  = os.path.join(datadir, "metadata/")
  handlers_file = os.path.join(metadata_dir, "mail-handlers.txt")
  if not os.path.exists(handlers_file):
    os.makedirs(metadata_dir, exist_ok=True)
    with open(handlers_file, "w") as f:
      f.write(sample_mailhandlers_str)

  with open(config_file, "w") as f:
    f.write(sample_config_str.replace("YOUR NAME", myname).replace("USER@DOMAIN.COM", myaddr).replace("~/bomail", datadir))

  print(setup_success_str)
  exit(0)


# Function to open and parse config file
def parse_config_file():
  options_dict = {}
  def parse_line(line):
    line = line.strip()
    if len(line) > 0 and line[0] != "#":
      if "=" in line:
        left, right = line.split("=")
        options_dict[left.strip()] = right.strip()
      else:
        options_dict[line.strip()] = True
  
  with open(config_file) as f:
    for line in f:
      parse_line(line)
  return options_dict


# Actually do it
options_dict = parse_config_file()


# --------------------------------------------------------------
# User info

name = options_dict["name"]
email_addr = options_dict["email_addr"]
my_aliases = []
if "alias_addresses" in options_dict:
  my_aliases = [s.strip() for s in options_dict["alias_addresses"].split(",")]

smtp_servername = options_dict["smtp_servername"]
smtp_port = int(options_dict["smtp_port"])
password = ""
if "smtp_password" in options_dict:
  password = options_dict["smtp_password"]
else:
  try:
    with open(os.path.join(home, ".getmail/getmailrc")) as f:
      for line in f:
        if line.startswith("password ="):
          password = line[len("password ="):].strip()
          break
  except:
    pass


# --------------------------------------------------------------
# Organization locations

def add_slash(s):
  return s if s[-1] == "/" else s + "/"
bomail_data_base = add_slash(options_dict["bomail_location"].replace("~", home))
new_rawmail_dir = add_slash(options_dict["new_rawmail_location"].replace("~", home))
old_rawmail_dir = add_slash(options_dict["processed_rawmail_location"].replace("~", home))


# --------------------------------------------------------------
# Options not in bomailrc (may edit)

# how many new emails to process at a time, -1 for no limit
process_new_limit = -1

# how many new emails to process in a batch, decrease to save memory
# for very large imports
process_batch_size = 20000

# tag drafts with same tags as parent
autotag_draft_replies = True

# Replying to emails
# will replace %from, %to, %date, %time, %body
# with the corresponding text from the quoted email
quote_format = """\n
%from wrote on %date at %time:
%body"""

# %a is day of week; %Y,%m,%d are year/month/date
# %H,%M are hour/minute
quote_date_fmt = "%a, %Y-%m-%d"  
quote_time_fmt = "%H:%M"         
quote_line_prefix = "> "


# --------------------------------------------------------------
# File locations: suggested not to edit

# Locations of directories
email_dir     = os.path.join(bomail_data_base, "email/")
attach_dir    = os.path.join(bomail_data_base, "attachments/")
drafts_dir    = os.path.join(bomail_data_base, "drafts/")
trash_dir     = os.path.join(bomail_data_base, "trash/")
metadata_dir  = os.path.join(bomail_data_base, "metadata/")

# Locations of files that must be edited by hand
handlers_file      = os.path.join(metadata_dir, "mail-handlers.txt")
addr_alias_file    = os.path.join(metadata_dir, "aliases.txt")

# Locations of files that may be edited by hand
tags_file          = os.path.join(metadata_dir, "tags.txt")
addr_book_file     = os.path.join(metadata_dir, "addr_book.txt")

# Locations of files that should not be hand-edited
openlist_file      = os.path.join(metadata_dir, "openlist.txt")
scheduledlist_file = os.path.join(metadata_dir, "scheduledlist.txt")
msg_ids_file       = os.path.join(metadata_dir, "msg_ids.txt")
tab_config_file    = os.path.join(metadata_dir, "tab_config.py")

# Locations of logs, these can be deleted whenever
acts_log_file      = os.path.join(metadata_dir, "acts_log.txt")
error_log_file     = os.path.join(metadata_dir, "error_log.txt")

for d in [email_dir, attach_dir, drafts_dir, trash_dir, metadata_dir]:
  if not os.path.exists(d):
    os.makedirs(d)





from setuptools import setup, find_packages
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

with open(path.join(here, "README.txt"), encoding="utf-8") as f:
  long_description = f.read()

setup(
  name = "bomail",
  version = "0.9.0.1",
  description = "A homemade, hobbyist, hacky system for organizing and interacting with email.",
  long_description = long_description,
  url = "https://bitbucket.org/bowaggonerpublic/bomail",
  author = "Bo Waggoner",
  author_email = "luv2runxc@gmail.com",
  install_requires = ["datetime", "python-dateutil"],
  python_requires = ">=3.3",
  packages = find_packages(exclude=['contrib', 'docs', 'tests']), #["bomail"],
  entry_points = {
    "console_scripts": ["bomail = bomail.bomail:main"]
  }
  )  # end setup




####
# bomail.cli.chstate
#
# Change state of files (open/closed/scheduled/trash).
####

import sys
import shutil
import os, subprocess, datetime
import tempfile
import dateutil.parser
from dateutil import tz
from dateutil.relativedelta import relativedelta
from subprocess import Popen, PIPE

import bomail.config.config as config
import bomail.cli.mailfile as mailfile
import bomail.util.remove_lines as remove_lines
import bomail.util.merge_lines as merge_lines
import bomail.util.util as util
import bomail.util.datestr as datestr

usage_str = """
Change the state of an email to open, closed, or scheduled.
Reads a list of filenames from stdin, one per line, and applies the operation.
Call with no args or -h to print this help.

  bomail chstate -o           # make filenames open
  bomail chstate -c           # make filenames closed
  bomail chstate -t           # trash filenames
  bomail chstate -ut          # untrash filenames
  bomail chstate -s datestr   # schedule filename for datestr (e.g. yyyy-mm-dd)
"""


# remove files from open and scheduled lists
def remove_from_lists(filelist):
  if os.path.exists(config.openlist_file):
    remove_lines.do_sorted(config.openlist_file, sorted(filelist))
  if os.path.exists(config.scheduledlist_file):
    remove_lines.do_sched(config.scheduledlist_file, set(filelist))


# change all files' state to open
def make_open(filelist, mgr, remove_old=True):
  if remove_old:
    remove_from_lists(filelist)
  # sort openlist by lexicographic
  merge_lines.do(config.openlist_file, sorted(filelist))
  for f in filelist:
    mgr.set(f, mailfile.STATE_L, "open")


# change all files' state to closed
def make_closed(filelist, mgr):
  remove_from_lists(filelist)
  for f in filelist:
    mgr.set(f, mailfile.STATE_L, "closed")


def scheduled_list_str(filename, sched_datestr):
  return sched_datestr + " " + filename


# schedule all files for sched
# sched is a datetime object
def schedule(filelist, sched, mgr):
  remove_from_lists(filelist)
  # round down to the minute
  rounded_date = sched - relativedelta(seconds=sched.second, microseconds=sched.microsecond)
  datestr = rounded_date.strftime("%Y-%m-%dT%H:%M")
  # sort scheduledlist
  lines = [scheduled_list_str(f, datestr) for f in filelist]
  merge_lines.do(config.scheduledlist_file, lines)
  for f in filelist:
    mgr.set(f, mailfile.STATE_L, "scheduled " + datestr)


# trash all files, including remove them from msg_ids_file
# return corresponding list of new filenames (in trash folder)
def trash(filelist, mgr):
  remove_from_lists(filelist)
  if not os.path.exists(config.trash_dir):
    os.makedirs(config.trash_dir)
  newnames = []
  for f in filelist:
    d = mgr.get_all(f)
    new_dir,new_ext = mailfile.get_trashfilepath(f, d)
    new_name = new_dir + new_ext
    os.makedirs(new_dir, exist_ok=True)
    if os.path.exists(f):
      util.mv_file(f, new_name)
    else:
      mailfile.write_to_file(new_name, d)
    if len(d[mailfile.ATTACH_L].strip()) > 0:
      attach_dir = mailfile.get_attach_dir(d)
      if os.path.exists(attach_dir):
        trash_attach_dir = mailfile.get_trash_attach_dir(f, d)
        os.makedirs(trash_attach_dir, exist_ok=True)
        for attach_f in os.listdir(attach_dir):
          util.mv_file(attach_dir + attach_f, trash_attach_dir + attach_f)
    newnames.append(new_name)
  mgr.trash(filelist)
  return newnames


def do_untrash(trashed_filenames, mgr):
  datalist = [mailfile.file_to_data(f) for f in trashed_filenames]
  create_filelist = ["".join(mailfile.get_mailfilepath(d)) if f[-4:] == "mail" else "".join(mailfile.get_draftfilepath(d)) for d,f in zip(datalist, trashed_filenames)]
  untrash(create_filelist, datalist, mgr)


# given list of new filenames and data to put in them
def untrash(create_filelist, datalist, mgr):
  # 1) create mailfiles
  create(create_filelist, datalist, mgr)
  # 2) move attachments back if needed
  for fname,d in zip(create_filelist,datalist):
    if len(d[mailfile.ATTACH_L].strip()) > 0:
      trash_attach_path = mailfile.get_trash_attach_dir(fname, d)
      dest_path = mailfile.get_attach_dir(d)
      if os.path.exists(dest_path):
        shutil.rmtree(dest_path)
      util.mv_file(trash_attach_path, dest_path)
    

# the opposite of trash basically
def create(filelist, datalist, mgr):
  for fname,d in zip(filelist, datalist):
    if os.path.exists(fname):
      with open(config.error_log_file, "a") as f:
        f.write("Creating file, already exists: " + fname + "\n")
        d2 = mailfile.file_to_data(fname)
        for i in range(len(d)):
          if d2[i] != d[i]:
            f.write(">>>> old version " + mailfile.fields[i] + ":\n")
            f.write(d2[i] + "\n")
            f.write(">>>> new version:\n")
            f.write(d[i] + "\n")
        f.write("\n")
    elif d[mailfile.MSG_ID_L] in mgr.ids.ids_to_filenames:
      with open(config.error_log_file, "a") as f:
        f.write("Creating file, non-identical file with same message ID already exists: " + mgr.ids.get(d[mailfile.MSG_ID_L]) + "\n")
        f.write("New data:\n  ")
        f.write("\n  ".join(d) + "\n")
    else:
      mailfile.write_to_file(fname, d)
  mgr.create(filelist, datalist)
  make_open(filelist, mgr, False)  # False == don't try to remove old ones from list
 

def main(args, filelist, mgr):
  filelist.sort()
  if args[0] == "-o":
    make_open(filelist, mgr)
  elif args[0] == "-c":
    make_closed(filelist, mgr)
  elif args[0] == "-t":
    trash(filelist, mgr)
  elif args[0] == "-ut":
    do_untrash(filelist, mgr)  # filelist are all currently in trash
  elif args[0] == "-s":
    sched = datestr.get_datetime(args[1])
    if sched is None:
      print("Could not parse scheduling date string!")
      exit(0)
    schedule(filelist, sched, mgr)


def main_cli():
  if len(sys.argv) < 2 or "-h" in sys.argv:
    print(usage_str)
    exit(0)
  mgr = mailfile.MailMgr()
  # remove each trailing newline
  main(sys.argv[1:], [s[:-1] for s in sys.stdin.readlines()], mgr)


if __name__ == "__main__":
  main_cli()




####
# bomail.config.guiconfig
#
# User Interface configuration file.
####

import curses

import bomail.config.config as config

options_dict = config.options_dict


# --------------------------------------------------------------
# "Important" UI options from config file

threads_on = False
if "threads_on" in options_dict:
  threads_on = True

read_prog = options_dict["read_program"]
edit_prog = options_dict["edit_program"]

arrowkey_nav = False
if "arrowkey_navigation" in options_dict:
  arrowkey_nav = True

colorscheme = options_dict["colorscheme"]


# --------------------------------------------------------------
# "Other" UI options from config file

remove_newlines = options_dict["strip_newlines_in_msg_preview"]

lines_per_msg = int(options_dict["total_lines_per_msg"])
thread_lines_per_collapsed_msg = int(options_dict["threadview_total_lines_per_msg"])

datelen = int(options_dict["date_len"])
datepad = int(options_dict["date_pad"])
authorlen = int(options_dict["author_len"])
authorpad = int(options_dict["author_pad"])

horiz_line = False
if "hline_between_msgs" in options_dict:
  horiz_line = True

skip_line = False
if "skipline_between_msgs" in options_dict:
  skip_line = True

show_tags = False
if "show_tags" in options_dict:
  show_tags = True

inline_tags = False
if "tags_on_topline" in options_dict:
  inline_tags = True


# --------------------------------------------------------------
# Options for threads, i.e. "conversations"

# Right now "linear" is the only option (TODO: add "tree")
thread_view = "linear"


# --------------------------------------------------------------
# Tab view spacing and formatting

# Add extra blank lines if message uses fewer than lines_per_msg
# (True or False)
pad_msg_lines = True

# Format for displaying dates
# Y = year, m = month, d = day, a = day of week, H = hour, M = minute
datefmt = "%Y-%m-%d %a %H:%M"

# formatting for shortened stuff
short_authorlen = 16
short_subjectlen = 32


# --------------------------------------------------------------
# Thread view spacing and formatting
# (Same as above, but for viewing messages in a given thread.)

thread_pad_msg_lines = pad_msg_lines
thread_horiz_line = horiz_line
thread_skip_line = skip_line
thread_remove_newlines = remove_newlines
thread_show_tags = show_tags
thread_inline_tags = inline_tags

# --------------------------------------------------------------
# Color formatting.
# Color options are COLOR_BLACK, COLOR_RED, COLOR_GREEN, COLOR_YELLOW,
# COLOR_BLUE, COLOR_MAGENTA, COLOR_CYAN, and COLOR_WHITE.

# Choose builtin color profile light1, light2, dark1, dark2
# or define custom colors.

if colorscheme == "light1":
  background_color = curses.COLOR_WHITE
  author_color = curses.COLOR_BLACK
  bomail_color = curses.COLOR_RED
  foreground_color = curses.COLOR_CYAN
  tags_color = curses.COLOR_MAGENTA
  thread_color = curses.COLOR_MAGENTA

elif colorscheme == "light2":
  background_color = curses.COLOR_WHITE
  author_color = curses.COLOR_RED
  bomail_color = curses.COLOR_BLACK
  foreground_color = curses.COLOR_MAGENTA
  tags_color = curses.COLOR_CYAN
  thread_color = curses.COLOR_CYAN

elif colorscheme == "dark1":
  background_color = curses.COLOR_BLACK
  author_color = curses.COLOR_WHITE
  bomail_color = curses.COLOR_GREEN
  foreground_color = curses.COLOR_CYAN
  tags_color = curses.COLOR_YELLOW
  thread_color = curses.COLOR_MAGENTA

elif colorscheme == "dark2":
  background_color = curses.COLOR_BLACK
  author_color = curses.COLOR_GREEN
  bomail_color = curses.COLOR_WHITE
  foreground_color = curses.COLOR_MAGENTA
  tags_color = curses.COLOR_CYAN
  thread_color = curses.COLOR_CYAN

# Custom color profile
# To define, uncomment (delete initial '#') and choose colors
#background_color = curses.COLOR_BLACK
#author_color = curses.COLOR_GREEN
#bomail_color = curses.COLOR_WHITE
#foreground_color = curses.COLOR_MAGENTA
#tags_color = curses.COLOR_CYAN
#thread_color = curses.COLOR_CYAN

# all color profiles share these attribute settings
bomail_attr  = curses.A_BOLD
author_attr  = 0  # nothing special for authors, currently
cursor_attr = curses.A_STANDOUT
quote_attr  = curses.A_DIM


# --------------------------------------------------------------
# Key bindings

QUIT_KEY      = "q"

UP_KEY        = "k"
DOWN_KEY      = "j"
WAY_UP_KEY    = "K"
WAY_DOWN_KEY  = "J"
RIGHT_KEY     = "l"
LEFT_KEY      = "h"
WAY_RIGHT_KEY = "L"
WAY_LEFT_KEY  = "H"

if arrowkey_nav:
  UP_KEY        = "KEY_UP"
  DOWN_KEY      = "KEY_DOWN"
  WAY_UP_KEY    = "KEY_SUP"    # note curses and terminals don't actually support this!
  WAY_DOWN_KEY  = "KEY_SDOWN"  # or this!
  RIGHT_KEY     = "KEY_RIGHT"
  LEFT_KEY      = "KEY_LEFT"
  WAY_RIGHT_KEY = "KEY_SRIGHT"
  WAY_LEFT_KEY  = "KEY_SLEFT"


MARK_KEY      = "x"
MARK_ALL_KEY  = "X"

VIEW_KEY      = "v"
EDIT_DRAFT_KEY = "e"

OPEN_KEY      = "o"
SCHEDULE_KEY  = "s"
CLOSE_KEY     = "c"
TRASH_KEY     = "t"

GET_KEY       = "g"
WRITE_KEY     = "w"
SEND_KEY      = "S"

ADD_TAGS_KEY  = "a"
REMOVE_TAGS_KEY = "d"

UNDO_KEY      = "u"
REDO_KEY      = "r"

TAB_LEFT_KEY  = "["
TAB_RIGHT_KEY = "]"
EDIT_TAB_KEY  = "e"
ADD_TAB_KEY   = "+"
REMOVE_TAB_KEY = "-"
MOVE_TAB_LEFT_KEY = "{"
MOVE_TAB_RIGHT_KEY = "}"

# keys in thread window
THREAD_QUIT_KEY = "q"
THREAD_EXPAND_KEY = "L"
THREAD_COLLAPSE_KEY = "H"
THREAD_TOGGLE_MSG_KEY = "e"
THREAD_MSG_UP_KEY = "K"
THREAD_MSG_DOWN_KEY = "J"
THREAD_ALL_OPEN_KEY = "O"
THREAD_ALL_SCHEDULE_KEY  = "S"
THREAD_ALL_CLOSE_KEY     = "C"
THREAD_ALL_TRASH_KEY     = "T"
THREAD_ALL_ADD_TAGS_KEY  = "A"
THREAD_ALL_REMOVE_TAGS_KEY = "D"


# --------------------------------------------------------------
# Don't change these, they are constants for formatting code.

DEFAULT_CLR_PAIR = 1
BOMAIL_CLR_PAIR = 2
AUTHOR_CLR_PAIR = 3
AUTHOR_HIGHLIGHT_CLR_PAIR = 4
TAGS_CLR_PAIR = 5
THREAD_CLR_PAIR = 6




####
# bomail.cli.check_sched
#
# Look through scheduledlist and put emails on openlist if their time has come.
####

import sys
import os
import tempfile

import bomail.config.config as config
import bomail.cli.chstate as chstate
import bomail.cli.mailfile as mailfile
import bomail.util.util as util


def main(mail_mgr):
  if not os.path.exists(config.scheduledlist_file):
    return []
  nowstr = util.local_now_timestr()
  to_open = []
  # read through scheduledlist_file opening emails until we find a date that exceeds now
  with open(config.scheduledlist_file) as sched_f, tempfile.NamedTemporaryFile(dir=".", mode="w", delete=False) as new_f:
    for full_line in sched_f:
      line = full_line.strip()
      if len(line) == 0:
        continue
      datestr, filename = line.split()
      if nowstr >= datestr:
        to_open.append(filename)
        # and don't write to new_f, and continue
      else:
        if len(to_open) == 0:
          break  # no changes to make, will discard new_f
        # need to write the rest of file to new_f
        new_f.write(full_line)
  if len(to_open) == 0:
    os.remove(new_f.name)
    return to_open
  util.mv_file(new_f.name, config.scheduledlist_file)
  chstate.make_open(to_open, mail_mgr)
  return to_open


def main_cli():
  changed_files = main(mailfile.MailMgr())
  if len(changed_files) == 0:
    print("No updates.")
  else:
    print("Opened " + str(len(changed_files)) + (" message:" if len(changed_files) == 1 else " messages:"))
    print("\n".join(changed_files))


if __name__ == "__main__":
  main_cli()




####
# bomail.cli.process
#
# Process any new emails and load them into bomail's formats, etc.
# Includes calling custom 'handlers' via bomail.util.handle.
####

import sys
import os, email
import time
from multiprocessing import Process, Queue

import bomail.config.config as config

import bomail.cli.chstate as chstate
import bomail.cli.mailfile as mailfile

import bomail.util.mime_to_mailfile as mime_to_mailfile
import bomail.util.unpack_attach as unpack_attach
import bomail.util.merge_lines as merge_lines
import bomail.util.handle as handle

from bomail.util.addr import AddrBook
from bomail.util.tags import TagMgr

####
# process new mail:
#  1. ensure all directories for mailfile, mime message, attachments exist
#  2. get email data - headers and plain text body
#  3. change to attachments dir, call munpack on mailfile, change back
#  4. save file, move to open list, and add to msg_ids database
#  5. move the mime message to cur/date/
#  6. update address book
#  7. handle all the files using handlers
####

usage_str = """Processes all newly arrived mail (in directory defined in config),
depositing it in cur/ (or as defined in config), creating bomail files, and
extracting attachments.

Options:
    -h           # print this help
    -limit n     # process at most n messages (-1 for unlimited, default is in config)
    -batch n     # process at most this many at a time (-1 for unlimited, default in config)

See also: config, mail-handlers.
"""


# speed up by not asking OS if a dir exists every time
# TODO test if this gives noticeable speedup
created_dirs = set()
def check_and_create_dir(name):
  if not name in created_dirs:
    created_dirs.add(name)
    os.makedirs(name, exist_ok=True)


# unpack attachments
def do_unpack(input_queue, output_queue):
  attach_prlist = []  # pairs (index, attach_list)
  ind = 0
  while True:
    pr = input_queue.get()  # blocks until something is put
    if pr is None:
      output_queue.put(attach_prlist)
      return
    # pr = (py_msg, attach_dir)
    mylist = unpack_attach.do_unpack(pr[0], pr[1])
    if len(mylist) > 0:
      attach_prlist.append((ind, mylist))
    ind += 1


def do_move(input_queue):
  while True:
    pr = input_queue.get()  # blocks until something is put
    if pr is None:
      return
    src,dest = pr   


def msg_id_filename(msg_id):
  return msg_id.replace("<","-").replace(">","-").replace("/","-")


# the main processing function
def process_list(newfilenames, mail_mgr, addr_book, handler, tag_mgr):
  # using this map ensures that a message-id fetched multiple times
  # is only created/processed once
  new_fname_to_data = {} 
  new_msgids = set()
  new_fnames = []
  rename_mimelist = []
  rm_redundant_mimelist = []

  unpack_in_q = Queue()
  unpack_out_q = Queue()
  unpack_process = Process(target=do_unpack, args=(unpack_in_q, unpack_out_q))
  unpack_process.start()
  
  try:
    for small_f in newfilenames:
      f = config.new_rawmail_dir + small_f
      with open(f, encoding='utf-8', errors='replace') as myfile:
        py_msg = email.message_from_file(myfile)
      maildata = mime_to_mailfile.get_data(py_msg)
    
      # 1. ensure directories exist
      mailfile_dir, mailfile_name = mailfile.get_mailfilepath(maildata)
      mailfile_path = mailfile_dir + mailfile_name
      msg_id = maildata[mailfile.MSG_ID_L]
      if os.path.exists(mailfile_path) or mailfile_path in new_fname_to_data or msg_id in mail_mgr.ids.ids_to_filenames or msg_id in new_msgids:
        # assume unique message_id / path means it's a duplicate!
        trash_dir, trash_ext = mailfile.get_trashfilepath(mailfile_path, maildata)
        trash_path = trash_dir + trash_ext
        rm_redundant_mimelist.append((f,trash_path))
        continue
      attach_dir = mailfile.get_attach_dir(maildata)
      rawdir, rawname = mailfile.get_rawfilepath(maildata)
      rename_mimelist.append((f, rawdir + rawname))
      check_and_create_dir(mailfile_dir)
      check_and_create_dir(rawdir)
      
      # 2. get email data
      new_fname_to_data[mailfile_path] = maildata
      new_fnames.append(mailfile_path)
      new_msgids.add(msg_id)
      
      # 3. unpack attachments (spawn process?)
      unpack_in_q.put((py_msg, attach_dir))
  finally:
    unpack_in_q.put(None)
    attach_prlist = unpack_out_q.get()
    for (i,l) in attach_prlist:
      new_fname_to_data[new_fnames[i]][mailfile.ATTACH_L] = ", ".join(l)


  # 4. create file, add to open list, and add to msg_ids file (also load into mail_mgr)
  pairs = new_fname_to_data.items()
  mailfiles = [p[0] for p in pairs]
  maildatas = [p[1] for p in pairs]
  chstate.create(mailfiles, maildatas, mail_mgr)

  # 5. update address book
  addr_book.update_for_new(maildatas)
  
  # 6. handle the mail
  handler.handle(mailfiles, mail_mgr)

  # 7. move mime files
  for a,b in rename_mimelist:
    os.rename(a, b)
  for a,b in rm_redundant_mimelist:
    check_and_create_dir(os.path.dirname(b))
    os.rename(a, b)

  return mailfiles


def main(mail_mgr, tag_mgr=None, process_new_limit=config.process_new_limit, batch_size=config.process_batch_size):
  
  # for each mime file:
  try:
    newfilenames = os.listdir(config.new_rawmail_dir)
  except:
    with open(config.error_log_file, "a") as f:
      f.write("Could not find new-email directory (" + config.new_rawmail_dir + ")")
    return []
  addr_book = AddrBook()
  handler = handle.MailHandler()
  tag_mgr = tag_mgr if tag_mgr is not None else TagMgr()
  if process_new_limit > 0:
    newfilenames = newfilenames[:process_new_limit]

  if batch_size <= 0:
   return process_list(newfilenames, mail_mgr, addr_book, handler, tag_mgr)
  
  mailfiles = []
  for i in range(0,len(newfilenames),batch_size):
    mailfiles += process_list(newfilenames[i:i+batch_size], mail_mgr, addr_book, handler, tag_mgr)
      
  return mailfiles


def main_cli():
  try:
    args = sys.argv[1:]
    if "-h" in args:
      raise
    process_limit = config.process_new_limit
    batch_size = config.process_batch_size
    if "-limit" in args:
      i = args.index("-limit")
      if i == len(args)-1:
        raise
      process_limit = int(args[i+1])
      args = args[:i] + ([] if i+2 >= len(args) else args[i+2:])
    if "-batch" in args:
      i = args.index("-batch")
      if i == len(args)-1:
        raise
      batch_size = int(args[i+1])
      args = args[:i] + ([] if i+2 >= len(args) else args[i+2:])
    if len(args) > 0:
      raise

    files = main(mailfile.MailMgr(), None, process_limit, batch_size)
    if len(files) == 0:
      print("No new messages.\n")
    else:
      print_files = len(files) <= 50      # number 50 chosen through extensive user studies
      print("Processed " + str(len(files)) + (" message" if len(files) == 1 else " messages") + (":" if print_files else "."))
      if print_files:
        print("\n".join(files))
  except:
    print(usage_str)


if __name__ == "__main__":
  main_cli()




####
# bomail.guistuff.tabthreadview
#
# The view of a tab when you navigate into
# a specific thread to see its list of
# emails.
####

import subprocess
import curses

import bomail.config.guiconfig as guicfg
import bomail.cli.mailfile as mailfile
import bomail.guistuff.display_fmt as display_fmt
import bomail.guistuff.gui_util as gui_util
import bomail.util.util as util


indic = lambda t: 1 if t else 0


class ThreadView:
  def __init__(self, trip, tab, gui):
    self.tab = tab
    self.gui = gui
    self.trip = trip
    self.matching_set = set(trip[0])  # files in thread that match search string
    self.repr_file = trip[0][0]       # representative file

    self.my_key = lambda f: self.gui.mail_mgr.get(f, mailfile.DATE_L)
    self.my_tup_key = lambda tup: self.my_key(tup[0])
    self.compare = lambda tup1, tup2: self.my_tup_key(tup1) <= self.my_tup_key(tup2)
    self.all_files = trip[2]          # all files in thread, chronological order
    self.all_files.sort(key=self.my_key)
    
    # filename, is_marked, is_collapsed, collapsed_disp_data, expanded_disp_data
    self.file_data = [[f, False, True, None, None] for f in self.all_files]

    self.is_display_loaded = False


  def ensure_display_loaded(self):
    if self.is_display_loaded:
      return
    self.top_y_coord = 0
    self.num_marked = 0
    self.cursor_ind = -1
    orig_data = self.gui.mail_mgr.get_all(self.all_files[0])  # use in display
    thread_total = len(self.file_data)
    ht, screen_width = self.gui.tab_area.getmaxyx()
    for i in range(len(self.file_data)):
      data = self.gui.mail_mgr.get_all(self.file_data[i][0])
      is_draft = self.file_data[i][0][-5:] == "draft"
      self.file_data[i][3] = display_fmt.get_thread_view_lines(data, orig_data, is_draft, screen_width, True, i+1, thread_total, 0)
      self.file_data[i][4] = display_fmt.get_thread_view_lines(data, orig_data, is_draft, screen_width, False, i+1, thread_total, 0)
      if self.file_data[i][0] in self.matching_set:
        self.file_data[i][2] = False  # expand
        if self.cursor_ind == -1:
          self.cursor_ind = i
    if self.cursor_ind == -1:
      self.cursor_ind = 0
    cursor_y = self.get_y_coord(self.cursor_ind)
    cursor_expanded_disp_pr = self.file_data[self.cursor_ind][4]
    last_visible_line = cursor_y + len(cursor_expanded_disp_pr[0])
    height,width = self.gui.tab_area.getmaxyx()
    first_visible_line = min(cursor_y, last_visible_line - height)
    self.top_y_coord = min(-first_visible_line, 0)

    self.is_display_loaded = True


  def get_curr_filenames(self):
    if len(self.file_data) == 0:
      return []
    elif self.num_marked == 0:
      return [self.file_data[self.cursor_ind][0]]
    else:
      return [t[0] for t in self.file_data if t[2]]  # marked


  # Attempt to reuse display data
  # Note filenames in old_threadview may not exist any more!
  def reuse_disp_data(self, old_threadview):
    if old_threadview.is_display_loaded == False:
      return  # nothing to reuse
    self.ensure_display_loaded()

    self.top_y_coord = old_threadview.top_y_coord
    i_new, i_old = 0, 0
    while i_new < len(self.file_data):
      if i_old >= len(old_threadview.file_data):
        break
      t_old = old_threadview.file_data[i_old]
      t_new = self.file_data[i_new]
      if t_old < t_new:  # old file no longer present
        i_old += 1
      elif t_old == t_new:  # old file still present
        t_new[1] = t_old[1]  # marked
        t_new[2] = t_old[2]  # collapsed
        if i_old == old_threadview.cursor_ind:
          self.cursor_ind = i_new
        i_old += 1
        i_new += 1
      else:  # new file
        t_new[1] = False
        t_new[2] = t_new[0] not in self.matching_set  # expand if matching
        i_new += 1
    self.num_marked = sum(t[2] for t in self.file_data)
    self.scroll_up_file(0)   # force cursor file into view
    self.scroll_down_file(0)
    self.scroll_up_line(0)
    self.scroll_down_line(0)


  def get_disp_prs(self, filename, orig_data, ind, num):
    data = self.gui.mail_mgr.get_all(filename)
    height, width = self.gui.screen.getmaxyx()
    pr_collapsed = display_fmt.get_thread_view_lines(data, orig_data, filename[-5:] == "draft", width, True, ind+1, num, 0)
    pr_expanded = display_fmt.get_thread_view_lines(data, orig_data, filename[-5:] == "draft", width, False, ind+1, num, 0)
    return pr_collapsed, pr_expanded


  def get_old_cursor_display(self):
    return self.file_data[self.cursor_ind][0], self.top_y_coord


  def update_display(self, old_cursor_file, old_top_y):
    old_date = self.gui.mail_mgr.get(old_cursor_file, mailfile.DATE_L)
    self.cursor_ind = util.bisect_left_key(self.file_data, old_date, key=self.my_tup_key)
    self.top_y_coord = old_top_y
    self.scroll_up_line(amt=0)
    self.scroll_down_line(amt=0)


  def get_y_coord(self, file_ind):
    y = self.top_y_coord
    for i in range(file_ind):
      disp_pr = self.file_data[i][3 if self.file_data[i][2] else 4]
      y += len(disp_pr[0])
    return y


  def get_last_y_coord(self, file_ind):
    disp_pr = self.file_data[file_ind][3 if self.file_data[file_ind][2] else 4]
    return self.get_y_coord(file_ind) + len(disp_pr[0]) - 1

  
  def scroll_up_file(self, amt=1):
    self.cursor_ind -= amt
    if self.cursor_ind < 0:
      self.cursor_ind = 0
    y = self.get_y_coord(self.cursor_ind)
    if y < 0:
      self.top_y_coord -= y


  def scroll_down_file(self, amt=1):
    self.cursor_ind += amt
    max_ind = len(self.all_files) - 1
    if self.cursor_ind > max_ind:
      self.cursor_ind = max_ind
    y = self.get_y_coord(self.cursor_ind)
    height, width = self.gui.tab_area.getmaxyx()
    if y >= height:
      self.top_y_coord -= y


  def scroll_up_line(self, amt=1):
    self.top_y_coord += amt
    if self.top_y_coord > 0:
      self.top_y_coord = 0
    height, width = self.gui.tab_area.getmaxyx()
    if self.get_y_coord(self.cursor_ind) >= height:
      self.cursor_ind -= 1


  def scroll_down_line(self, amt=1):
    self.top_y_coord -= amt
    height, width = self.gui.tab_area.getmaxyx()
    if self.cursor_ind < len(self.file_data)-1 and self.get_y_coord(self.cursor_ind) < 0 and self.get_y_coord(self.cursor_ind+1) < height:
      self.cursor_ind += 1
    if self.cursor_ind == len(self.file_data)-1:
      self.top_y_coord -= self.get_last_y_coord(self.cursor_ind)


  def collapse_current(self):
    self.file_data[self.cursor_ind][2] = True
    my_y = self.get_y_coord(self.cursor_ind)
    if my_y < 0:
      self.top_y_coord -= my_y


  def collapse_all(self):
    for tup in self.file_data:
      tup[2] = True
    my_y = self.get_y_coord(self.cursor_ind)
    if my_y < 0:
      self.top_y_coord -= my_y


  def expand_current(self):
    self.file_data[self.cursor_ind][2] = False


  def expand_all(self):
    for tup in self.file_data:
      tup[2] = False
    height, width = self.gui.tab_area.getmaxyx()
    my_y = self.get_y_coord(self.cursor_ind)
    if my_y >= height:
      self.top_y_coord -= (my_y - height)


  def process_key(self, key):
    self.ensure_display_loaded()
    kind, note = "note", ""
    if key == guicfg.DOWN_KEY:
      self.scroll_down_file(1)
      kind, note = "tab", ""
    elif key == guicfg.UP_KEY:
      self.scroll_up_file(1)
      kind, note = "tab", ""
    elif key == guicfg.WAY_DOWN_KEY:
      self.scroll_down_line(2)
      kind, note = "tab", ""
    elif key == guicfg.WAY_UP_KEY:
      self.scroll_up_line(2)
      kind, note = "tab", ""
    elif key == guicfg.RIGHT_KEY:
#    elif (guicfg.arrowkey_nav and key == "KEY_RIGHT") or (not guicfg.arrowkey_nav and key == guicfg.RIGHT_KEY):
      self.expand_current()
      kind, note = "tab", ""
    elif key == guicfg.LEFT_KEY:
#    elif (guicfg.arrowkey_nav and key == "KEY_LEFT") or (not guicfg.arrowkey_nav and key == guicfg.LEFT_KEY):
      # if collapsed already, then go back
      if self.file_data[self.cursor_ind][2]:
        self.tab.gui.mark_read(self.all_files)
        self.tab.mode = "thread list"
      else:
        self.collapse_current()
      kind, note = "tab", ""
    elif key == guicfg.WAY_RIGHT_KEY:
      self.expand_all()
      kind, note = "tab", ""
    elif key == guicfg.WAY_LEFT_KEY:
      self.collapse_all()
      kind, note = "tab", ""

    elif key == guicfg.MARK_KEY:
      summary = display_fmt.get_shortened(self.file_data[self.cursor_ind][0], self.gui.mail_mgr)
      if self.file_data[self.cursor_ind][1]:
        self.file_data[self.cursor_ind][1] = False
        self.num_marked -= 1
        kind, note = "tab", "Un-marked " + summary
      else:
        self.file_data[self.cursor_ind][1] = True
        self.num_marked += 1
        kind, note = "tab", "Marked " + summary
    elif key == guicfg.MARK_ALL_KEY:
      if self.num_marked == len(self.all_files):
        for tup in self.file_data:
          tup[1] = False
        self.num_marked = 0
        kind, note = "tab", "Un-marked all"
      else:
        for tup in self.file_data:
          tup[1] = True
        self.num_marked = len(self.all_files)
        kind, note = "tab", "Marked all"

    # writing and viewing
    elif key == guicfg.WRITE_KEY:
      filename = self.file_data[self.cursor_ind][0]
      mode, note = gui_util.go_write_key(self.gui, filename)

    elif key == guicfg.VIEW_KEY:
      self.gui.mark_read([self.file_data[self.cursor_ind][0]])
      subprocess.call(guicfg.read_prog + " '" + self.file_data[self.cursor_ind][0] + "'", shell=True)
      curses.curs_set(False)
      return "all", ""

    # sending
    elif key == guicfg.SEND_KEY:
      filename = self.file_data[self.cursor_ind][0]
      if filename[-5:] == "draft":
        return gui_util.go_send(self.gui, filename)
      else:
        return "note", "Cannot send: not a draft"

    else:
      return "note", "Input key not recognized"
    return kind, note




####
# bomail.cli.gui
#
# Launch the user interface (which is text based, so not a GUI. I know)
####

import sys
import os
import locale
import curses

import bomail.guistuff.mygui as mygui


def main(screen):
  mygui.main(screen)


def main_cli():
  os.environ.setdefault('ESCDELAY', '25')
  locale.setlocale(locale.LC_ALL, '')
  curses.wrapper(main)


if __name__ == "__main__":
  main_cli()

